#!/usr/bin/env python
'''param command handling'''

import time, os, fnmatch, struct, sys
from pymavlink import mavutil, mavparm
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import param_help
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

try:
    # py2
    from StringIO import StringIO as SIO
except ImportError:
    # py3
    from io import BytesIO as SIO
    
class ParamState:
    '''this class is separated to make it possible to use the parameter
       functions on a secondary connection'''
    def __init__(self, mav_param, logdir, vehicle_name, parm_file, mpstate, sysid):
        self.mav_param_set = set()
        self.mav_param_count = 0
        self.param_period = mavutil.periodic_event(1)
        self.fetch_one = dict()
        self.mav_param = mav_param
        self.logdir = logdir
        self.vehicle_name = vehicle_name
        self.parm_file = parm_file
        self.fetch_set = None
        self.new_sysid_timestamp = time.time()
        self.autopilot_type_by_sysid = {}
        self.param_types = {}
        self.ftp_failed = False
        self.ftp_started = False
        self.ftp_count = None
        self.ftp_send_param = None
        self.mpstate = mpstate
        self.sysid = sysid
        self.param_help = param_help.ParamHelp()
        self.param_help.vehicle_name = vehicle_name

    def use_ftp(self):
        '''return true if we should try ftp for download'''
        if self.ftp_failed:
            return False
        return self.mpstate.settings.param_ftp

    def handle_px4_param_value(self, m):
        '''special handling for the px4 style of PARAM_VALUE'''
        if m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_REAL32:
            # already right type
            return m.param_value
        is_px4_params = False
        if m.get_srcComponent() in [mavutil.mavlink.MAV_COMP_ID_UDP_BRIDGE]:
            # ESP8266 uses PX4 style parameters
            is_px4_params = True
        sysid = m.get_srcSystem()
        if self.autopilot_type_by_sysid.get(sysid,-1) in [mavutil.mavlink.MAV_AUTOPILOT_PX4]:
            is_px4_params = True
        if not is_px4_params:
            return m.param_value
        # try to extract px4 param value
        value = m.param_value
        try:
            v = struct.pack(">f", value)
        except Exception:
            return value
        if m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_UINT8:
            value, = struct.unpack(">B", v[3:])
        elif m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_INT8:
            value, = struct.unpack(">b", v[3:])
        elif m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_UINT16:
            value, = struct.unpack(">H", v[2:])
        elif m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_INT16:
            value, = struct.unpack(">h", v[2:])
        elif m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_UINT32:
            value, = struct.unpack(">I", v[0:])
        elif m.param_type == mavutil.mavlink.MAV_PARAM_TYPE_INT32:
            value, = struct.unpack(">i", v[0:])
        # can't pack other types

        # remember type for param set
        self.param_types[m.param_id.upper()] = m.param_type
        return value

    def handle_mavlink_packet(self, master, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'PARAM_VALUE':
            value = self.handle_px4_param_value(m)
            param_id = "%.16s" % m.param_id
            # Note: the xml specifies param_index is a uint16, so -1 in that field will show as 65535
            # We accept both -1 and 65535 as 'unknown index' to future proof us against someday having that
            # xml fixed.
            if self.fetch_set is not None:
                self.fetch_set.discard(m.param_index)
            if m.param_index != -1 and m.param_index != 65535 and m.param_index not in self.mav_param_set:
                added_new_parameter = True
                self.mav_param_set.add(m.param_index)
            else:
                added_new_parameter = False
            if m.param_count != -1:
                self.mav_param_count = m.param_count
            self.mav_param[str(param_id)] = value
            if param_id in self.fetch_one and self.fetch_one[param_id] > 0:
                self.fetch_one[param_id] -= 1
                if isinstance(value, float):
                    print("%s = %.7f" % (param_id, value))
                else:
                    print("%s = %s" % (param_id, str(value)))
            if added_new_parameter and len(self.mav_param_set) == m.param_count:
                print("Received %u parameters" % m.param_count)
                if self.logdir is not None:
                    self.mav_param.save(os.path.join(self.logdir, self.parm_file), '*', verbose=True)
                self.fetch_set = None
            if self.fetch_set is not None and len(self.fetch_set) == 0:
                self.fetch_check(master, force=True)
        elif m.get_type() == 'HEARTBEAT':
            if m.get_srcComponent() == 1:
                # remember autopilot types so we can handle PX4 parameters
                self.autopilot_type_by_sysid[m.get_srcSystem()] = m.autopilot

    def fetch_check(self, master, force=False):
        '''check for missing parameters periodically'''
        if self.param_period.trigger() or force:
            if master is None:
                return
            if len(self.mav_param_set) == 0 and not self.ftp_started:
                if not self.use_ftp():
                    master.param_fetch_all()
                else:
                    self.ftp_start()
            elif not self.ftp_started and self.mav_param_count != 0 and len(self.mav_param_set) != self.mav_param_count:
                if master.time_since('PARAM_VALUE') >= 1 or force:
                    diff = set(range(self.mav_param_count)).difference(self.mav_param_set)
                    count = 0
                    while len(diff) > 0 and count < 10:
                        idx = diff.pop()
                        master.param_fetch_one(idx)
                        if self.fetch_set is None:
                            self.fetch_set = set()
                        self.fetch_set.add(idx)
                        count += 1

    def param_use_xml_filepath(self, filepath):
        self.param_help.xml_filepath = filepath

    def param_set_xml_filepath(self, args):
        self.param_use_xml_filepath(args[0])

    def status(self, master, mpstate):
        return(len(self.mav_param_set), self.mav_param_count)

    def ftp_start(self):
        '''start a ftp download of parameters'''
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            self.ftp_failed = True
            self.ftp_started = False
            return
        self.ftp_started = True
        self.ftp_count = None
        ftp.cmd_get(["@PARAM/param.pck?withdefaults=1"], callback=self.ftp_callback, callback_progress=self.ftp_callback_progress)

    def log_params(self, params):
        '''log PARAM_VALUE messages so that we can extract parameters from a tlog when using ftp download'''
        if not self.mpstate.logqueue:
            return
        usec = int(time.time() * 1.0e6)
        idx = 0
        mav = self.mpstate.master().mav
        editor = self.mpstate.module('paramedit')
        for (name, v, ptype) in params:
            p = mavutil.mavlink.MAVLink_param_value_message(name,
                                                            float(v),
                                                            mavutil.mavlink.MAV_PARAM_TYPE_REAL32,
                                                            len(params),
                                                            idx)
            idx += 1
            # log PARAM_VALUE using the source vehicles sysid but MAV_COMP_ID_MISSIONPLANNER, to allow
            # us to tell that it came from the GCS, but to not corrupt the sequence numbers
            id_saved = (mav.srcSystem, mav.srcComponent)
            mav.srcSystem = self.sysid[0]
            mav.srcComponent = mavutil.mavlink.MAV_COMP_ID_MISSIONPLANNER
            try:
                buf = p.pack(mav)
                self.mpstate.logqueue.put(bytearray(struct.pack('>Q', usec) + buf))
                # also give to param editor so it can update for changes
                if editor:
                    editor.mavlink_packet(p)
            except Exception:
                pass
            (mav.srcSystem, mav.srcComponent) = id_saved

    def ftp_callback_progress(self, fh, total_size):
        '''callback as read progresses'''
        if self.ftp_count is None and total_size >= 6:
            ofs = fh.tell()
            fh.seek(0)
            buf = fh.read(6)
            fh.seek(ofs)
            magic2,num_params,total_params = struct.unpack("<HHH", buf)
            if magic2 == 0x671b or magic2 == 0x671c:
                self.ftp_count = total_params
        # approximate count
        if self.ftp_count is not None:
            # each entry takes 11 bytes on average
            per_entry_size = 11
            done = min(int(total_size / per_entry_size), self.ftp_count-1)
            self.mpstate.console.set_status('Params', 'Param %u/%u' % (done, self.ftp_count))

    def ftp_callback(self, fh):
        '''callback from ftp fetch of parameters'''
        self.ftp_started = False
        if fh is None:
            # the fetch failed
            self.ftp_failed = True
            return

        magic = 0x671b
        magic_defaults = 0x671c
        data = fh.read()
        magic2,num_params,total_params = struct.unpack("<HHH", data[0:6])
        if magic != magic2 and magic_defaults != magic2:
            print("paramftp: bad magic 0x%x expected 0x%x" % (magic2, magic))
            return
        with_defaults = magic2 == magic_defaults
        data = data[6:]

        # mapping of data type to type length and format
        data_types = {
            1: (1, 'b'),
            2: (2, 'h'),
            3: (4, 'i'),
            4: (4, 'f'),
        }

        count = 0
        params = []
        if with_defaults:
            defaults = []

        if sys.version_info.major < 3:
            pad_byte = chr(0)
            last_name = ''
        else:
            pad_byte = 0
            last_name = bytes()

        while True:
            while len(data) > 0 and data[0] == pad_byte:
                # skip pad bytes
                data = data[1:]

            if len(data) == 0:
                break

            ptype, plen = struct.unpack("<BB", data[0:2])
            flags = (ptype>>4) & 0x0F
            has_default = with_defaults and (flags&1) != 0
            ptype &= 0x0F


            if not ptype in data_types:
                print("paramftp: bad type 0x%x" % ptype)
                return

            (type_len, type_format) = data_types[ptype]
            default_len = type_len if has_default else 0

            name_len = ((plen>>4) & 0x0F) + 1
            common_len = (plen & 0x0F)
            name = last_name[0:common_len] + data[2:2+name_len]
            vdata = data[2+name_len:2+name_len+type_len+default_len]
            last_name = name
            data = data[2+name_len+type_len+default_len:]
            if with_defaults:
                if has_default:
                    v1,v2, = struct.unpack("<" + type_format + type_format, vdata)
                    params.append((name, v1, ptype))
                    defaults.append((name, v2, ptype))
                else:
                    v, = struct.unpack("<" + type_format, vdata)
                    params.append((name, v, ptype))
                    defaults.append((name, v, ptype))
            else:
                v, = struct.unpack("<" + type_format, vdata)
                params.append((name, v, ptype))
            count += 1

        if count != total_params:
            print("paramftp: bad count %u should be %u" % (count, total_params))
            return
        self.param_types = {}
        self.mav_param_set = set()
        self.fetch_one = dict()
        self.fetch_set = None
        self.mav_param.clear()
        self.mav_param_count = total_params

        idx = 0
        for (name, v, ptype) in params:
            # we need to set it to REAL32 to ensure we use write value for param_set
            name = str(name.decode('utf-8'))
            self.param_types[name] = mavutil.mavlink.MAV_PARAM_TYPE_REAL32
            self.mav_param_set.add(idx)
            self.mav_param[name] = v
            idx += 1

        self.ftp_failed = False
        self.mpstate.console.set_status('Params', 'Param %u/%u' % (total_params, total_params))
        print("Received %u parameters (ftp)" % total_params)
        if self.logdir is not None:
            self.mav_param.save(os.path.join(self.logdir, self.parm_file), '*', verbose=True)
        self.log_params(params)

        if with_defaults:
            defaults_path = os.path.join(self.logdir, "defaults.parm") if self.logdir else "defaults.parm"
            defparm = mavparm.MAVParmDict()
            for (name, v, ptype) in defaults:
                name = str(name.decode('utf-8'))
                defparm[name] = v
            defparm.save(defaults_path, '*', verbose=False)
            print("Saved %u defaults to %s" % (len(defaults), defaults_path))



    def fetch_all(self, master):
        '''force refetch of parameters'''
        if not self.use_ftp():
            master.param_fetch_all()
            self.mav_param_set = set()
        else:
            self.ftp_start()

    def handle_command(self, master, mpstate, args):
        '''handle parameter commands'''
        param_wildcard = "*"
        usage="Usage: param <fetch|ftp|save|set|show|load|preload|forceload|ftpload|diff|download|check|help>"
        if len(args) < 1:
            print(usage)
            return
        if args[0] == "fetch":
            if len(args) == 1:
                self.fetch_all(master)
                if self.ftp_started:
                    print("Requested parameter list (ftp)")
                else:
                    print("Requested parameter list")
            else:
                found = False
                pname = args[1].upper()
                for p in self.mav_param.keys():
                    if fnmatch.fnmatch(p, pname):
                        master.param_fetch_one(p)
                        if p not in self.fetch_one:
                            self.fetch_one[p] = 0
                        self.fetch_one[p] += 1
                        found = True
                        print("Requested parameter %s" % p)
                if not found and args[1].find('*') == -1:
                    master.param_fetch_one(pname)
                    if pname not in self.fetch_one:
                        self.fetch_one[pname] = 0
                    self.fetch_one[pname] += 1
                    print("Requested parameter %s" % pname)
        elif args[0] == "ftp":
            self.ftp_start()

        elif args[0] == "save":
            if len(args) < 2:
                print("usage: param save <filename> [wildcard]")
                return
            if len(args) > 2:
                param_wildcard = args[2]
            else:
                param_wildcard = "*"
            self.mav_param.save(args[1].strip('"'), param_wildcard, verbose=True)
        elif args[0] == "diff":
            wildcard = '*'
            if len(args) < 2 or args[1].find('*') != -1:
                filename = os.path.join(self.logdir, "defaults.parm")
                if not os.path.exists(filename):
                    if self.vehicle_name is None:
                        print("Unknown vehicle type")
                        return
                    filename = mp_util.dot_mavproxy("%s-defaults.parm" % self.vehicle_name)
                    if not os.path.exists(filename):
                        print("Please run 'param download' first (vehicle_name=%s)" % self.vehicle_name)
                        return
                if len(args) >= 2:
                    wildcard = args[1]
            else:
                filename = args[1]
                if len(args) == 3:
                    wildcard = args[2]
            print("defaults path: %s" % filename)
            print("%-16.16s %12.12s %12.12s" % ('Parameter', 'Defaults', 'Current'))
            self.mav_param.diff(filename, wildcard=wildcard)
        elif args[0] == "set":
            if len(args) < 2:
                print("Usage: param set PARMNAME VALUE")
                return
            if len(args) == 2:
                self.param_show(args[1], self.mpstate.settings.param_docs)
                return
            param = args[1]
            value = args[2]
            if value.startswith('0x'):
                value = int(value, base=16)
            if not param.upper() in self.mav_param:
                print("Unable to find parameter '%s'" % param)
                return
            uname = param.upper()
            ptype = None
            if uname in self.param_types:
                ptype = self.param_types[uname]
            self.mav_param.mavset(master, uname, value, retries=3, parm_type=ptype)

            if (param.upper() == "WP_LOITER_RAD" or param.upper() == "LAND_BREAK_PATH"):
                #need to redraw rally points
                mpstate.module('rally').set_last_change(time.time())
                #need to redraw loiter points
                mpstate.module('wp').wploader.last_change = time.time()

        elif args[0] == "load":
            if len(args) < 2:
                print("Usage: param load <filename> [wildcard]")
                return
            if len(args) > 2:
                param_wildcard = args[2]
            else:
                param_wildcard = "*"
            self.mav_param.load(args[1].strip('"'), param_wildcard, master)
        elif args[0] == "preload":
            if len(args) < 2:
                print("Usage: param preload <filename>")
                return
            self.mav_param.load(args[1].strip('"'))
        elif args[0] == "forceload":
            if len(args) < 2:
                print("Usage: param forceload <filename> [wildcard]")
                return
            if len(args) > 2:
                param_wildcard = args[2]
            else:
                param_wildcard = "*"
            self.mav_param.load(args[1].strip('"'), param_wildcard, master, check=False)
        elif args[0] == "ftpload":
            if len(args) < 2:
                print("Usage: param ftpload <filename> [wildcard]")
                return
            if len(args) > 2:
                param_wildcard = args[2]
            else:
                param_wildcard = "*"
            self.ftp_load(args[1].strip('"'), param_wildcard, master)
        elif args[0] == "download":
            self.param_help.param_help_download()
        elif args[0] == "apropos":
            self.param_help.param_apropos(args[1:])
        elif args[0] == "check":
            self.param_help.param_check(self.mav_param, args[1:])
        elif args[0] == "help":
            self.param_help.param_help(args[1:])
        elif args[0] == "set_xml_filepath":
            self.param_set_xml_filepath(args[1:])
        elif args[0] == "show":
            verbose = self.mpstate.settings.param_docs
            if len(args) > 1:
                pattern = args[1]
                if len(args) > 2 and args[2] == '-v':
                    verbose = True
            else:
                pattern = "*"
            self.param_show(pattern, verbose)
        elif args[0] == "status":
            print("Have %u/%u params" % (len(self.mav_param_set), self.mav_param_count))
        else:
            print(usage)

    def param_show(self, pattern, verbose):
        '''show parameters'''
        k = sorted(self.mav_param.keys())
        for p in k:
            name = str(p).upper()
            if fnmatch.fnmatch(name, pattern.upper()):
                value = self.mav_param.get(name)
                s = "%-16.16s %s" % (name, value)
                if verbose:
                    info = self.param_help.param_info(name, value)
                    if info is not None:
                        s = "%-28.28s # %s" % (s, info)
                print(s)

    def ftp_upload_callback(self, dlen):
        '''callback on ftp put completion'''
        if dlen is None:
            print("Failed to send parameters")
        else:
            if self.ftp_send_param is not None:
                for k in sorted(self.ftp_send_param.keys()):
                    v = self.ftp_send_param.get(k)
                    self.mav_param[k] = v
                self.ftp_send_param = None
            print("Parameter upload done")

    def ftp_upload_progress(self, proportion):
        '''callback from ftp put of parameters'''
        if proportion is None:
            self.mpstate.console.set_status('Params', 'Params ERR')
        else:
            self.mpstate.console.set_status('Params', 'Param %.1f%%' % (100.0*proportion))

    def best_type(self, v):
        '''work out best type for a variable'''
        if not v.is_integer():
            # treat as float
            return 4
        if v >= -128 and v <= 127:
            return 1
        if v >= -32768 and v <= 32767:
            return 2
        return 3

    def str_common_len(self, s1, s2):
        '''return common length between two strings'''
        c = min(len(s1),len(s2))
        for i in range(c):
            if s1[i] != s2[i]:
                return i
        return c

    def ftp_load(self, filename, param_wildcard, master):
        '''load parameters with ftp'''
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            print("Need ftp module")
            return
        newparm = mavparm.MAVParmDict()
        newparm.load(filename, param_wildcard, check=False)
        fh = SIO()
        for k in sorted(newparm.keys()):
            v = newparm.get(k)
            oldv = self.mav_param.get(k, None)
            if oldv is not None and abs(oldv - v) <= newparm.mindelta:
                # not changed
                newparm.pop(k)
        count = len(newparm.keys())
        if count == 0:
            print("No parameter changes")
            return

        fh.write(struct.pack("<HHH", 0x671b, count, count))
        last_param = ""
        for k in sorted(newparm.keys()):
            v = newparm.get(k)
            vtype = self.best_type(v)
            common_len = self.str_common_len(last_param, k)
            b1 = vtype
            b2 = common_len | ((len(k)-(common_len+1))<<4)
            fh.write(struct.pack("<BB", b1, b2))
            fh.write(k[common_len:].encode("UTF-8"))
            if vtype == 1: # int8
                fh.write(struct.pack("<b", int(v)))
            if vtype == 2: # int16
                fh.write(struct.pack("<h", int(v)))
            if vtype == 3: # int32
                fh.write(struct.pack("<i", int(v)))
            if vtype == 4: # float
                fh.write(struct.pack("<f", v))
            last_param = k

        # re-pack with full file length in total_params
        file_len = fh.tell()
        fh.seek(0)
        fh.write(struct.pack("<HHH", 0x671b, count, file_len))
        fh.seek(0)
        self.ftp_send_param = newparm
        print("Sending %u params" % count)
        ftp.cmd_put(["-", "@PARAM/param.pck"],
                    fh=fh, callback=self.ftp_upload_callback, progress_callback=self.ftp_upload_progress)


class ParamModule(mp_module.MPModule):
    def __init__(self, mpstate, **kwargs):
        super(ParamModule, self).__init__(mpstate, "param", "parameter handling", public = True, multi_vehicle=True)
        self.xml_filepath = kwargs.get("xml-filepath", None)
        self.pstate = {}
        self.check_new_target_system()
        self.menu_added_console = False
        self.add_command('param', self.cmd_param, "parameter handling",
                         ["<download|status>",
                          "<set|show|fetch|ftp|help|apropos> (PARAMETER)",
                          "<load|save|diff|forceload|ftpload> (FILENAME)",
                          "<set_xml_filepath> (FILEPATH)"
                         ])
        if mp_util.has_wxpython:
            self.menu = MPMenuSubMenu('Parameter',
                                  items=[MPMenuItem('Editor', 'Editor', '# module load paramedit'),
                                         MPMenuItem('Fetch', 'Fetch', '# param fetch'),
                                         MPMenuItem('Load', 'Load', '# param load ',
                                                    handler=MPMenuCallFileDialog(flags=('open',),
                                                                                 title='Param Load',
                                                                                 wildcard='ParmFiles(*.parm,*.param)|*.parm;*.param')),
                                         MPMenuItem('Save', 'Save', '# param save ',
                                                    handler=MPMenuCallFileDialog(flags=('save', 'overwrite_prompt'),
                                                                                 title='Param Save',
                                                                                 wildcard='ParmFiles(*.parm,*.param)|*.parm;*.param')),
                                         MPMenuItem('FTP', 'FTP', '# param ftp'),
                                         MPMenuItem('Update Metadata', 'Update Metadata', '# param download'),
                                             ])

    def get_component_id_list(self, system_id):
        '''get list of component IDs with parameters for a given system ID'''
        ret = []
        for (s,c) in self.mpstate.mav_param_by_sysid.keys():
            if s == system_id:
                ret.append(c)
        return ret

    def add_new_target_system(self, sysid):
        '''handle a new target_system'''
        if sysid in self.pstate:
            return
        if not sysid in self.mpstate.mav_param_by_sysid:
            self.mpstate.mav_param_by_sysid[sysid] = mavparm.MAVParmDict()
            self.new_sysid_timestamp = time.time()
        fname = 'mav.parm'
        if sysid not in [(0,0),(1,1),(1,0)]:

            fname = 'mav_%u_%u.parm' % (sysid[0], sysid[1])
        self.pstate[sysid] = ParamState(self.mpstate.mav_param_by_sysid[sysid], self.logdir, self.vehicle_name, fname, self.mpstate, sysid)
        if self.continue_mode and self.logdir is not None:
            parmfile = os.path.join(self.logdir, fname)
            if os.path.exists(parmfile):
                mpstate.mav_param.load(parmfile)
                self.pstate[sysid].mav_param_set = set(self.mav_param.keys())
        self.pstate[sysid].param_help.xml_filepath = self.xml_filepath

    def get_sysid(self):
        '''get sysid tuple to use for parameters'''
        component = self.target_component
        if component == 0:
            component = 1
        return (self.target_system, component)

    def check_new_target_system(self):
        '''handle a new target_system'''
        sysid = self.get_sysid()
        if sysid in self.pstate:
            return
        self.add_new_target_system(sysid)

    def param_status(self):
        sysid = self.get_sysid()
        pset, pcount = self.pstate[sysid].status(self.master, self.mpstate)
        return (pset, pcount)
        
    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        sysid = (m.get_srcSystem(),m.get_srcComponent())
        self.add_new_target_system(sysid)
        self.pstate[sysid].handle_mavlink_packet(self.master, m)

    def idle_task(self):
        '''handle missing parameters'''
        self.check_new_target_system()
        sysid = self.get_sysid()
        self.pstate[sysid].vehicle_name = self.vehicle_name
        self.pstate[sysid].param_help.vehicle_name = self.vehicle_name
        self.pstate[sysid].fetch_check(self.master)
        if self.module('console') is not None:
            if not self.menu_added_console:
                self.menu_added_console = True
                self.module('console').add_menu(self.menu)
        else:
            self.menu_added_console = False

    def cmd_param(self, args):
        '''control parameters'''
        self.check_new_target_system()
        sysid = self.get_sysid()
        self.pstate[sysid].handle_command(self.master, self.mpstate, args)

    def fetch_all(self):
        '''force fetch of all parameters'''
        self.check_new_target_system()
        sysid = self.get_sysid()
        self.pstate[sysid].fetch_all(self.master)


def init(mpstate, **kwargs):
    '''initialise module'''
    return ParamModule(mpstate, **kwargs)



#!/usr/bin/env python
#***************************************************************************
#                      Copyright 2017 Jaime Machuca
#***************************************************************************
# Title        : sc_ExifWriter.py
#
# Description  : This file is used to write the geotag informaton into the
#                EXIF metadata of a picture file. The file was originaly written
#                by Escadrone and distributed under Apache V2.0 license which is
#                retained by this file. This file may be modified from the original
#                in order to make it compatible with this MAVProxy Module.
#                The original file can be found here:
#
#                https://github.com/Escadrone/Solo-Mapper.git
#
# Environment  : Python 2.7 Code. Intended to be included in MAVproxy as a Module
#
# Responsible  : Jaime Machuca
#
# License      : Licensed under the Apache License, Version 2.0 (the "License");
#                you may not use this file except in compliance with the License.
#                You may obtain a copy of the License at
#
#                http://www.apache.org/licenses/LICENSE-2.0
#
#                Unless required by applicable law or agreed to in writing, software
#                distributed under the License is distributed on an "AS IS" BASIS,
#                WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#                See the License for the specific language governing permissions and
#                limitations under the License.
#
# Editor Used  : Xcode 6.1.1 (6A2008a)
#
#****************************************************************************

#****************************************************************************
# HEADER-FILES (Only those that are needed in this file)
#****************************************************************************

# System Header files and Module Headers

# Module Dependent Headers
import piexif

# Own Headers

#****************************************************************************
# LOCAL DEFINES
#****************************************************************************

#****************************************************************************
# Class name       : ExifWriter
#
# Public Methods   : write_gps
#
# Private Methods  : _decdeg2dms
#
#****************************************************************************

class ExifWriter:

#****************************************************************************
#   Method Name     : write_gps
#
#   Description     : Writes GPS coordinates into the given pictures EXIF
#
#   Parameters      : filename      Path to the image file
#                     lat           Latitude in decimal degrees to write
#                     longit        Longitude in decimal degrees to write
#                     altitude      Altitude above mean sea level in meters
#
#   Return Value    : None
#
#   Author           : Gael Billon, Escadrone
#
#****************************************************************************

    @staticmethod
    def write_gps(filename, lat, longit, altitude): #lat and long are decimals
        gps = {} #create array

        #Altitude
        if altitude<0:
            gps[piexif.GPSIFD.GPSAltitudeRef] = 1; #Below see level
        else:
            gps[piexif.GPSIFD.GPSAltitudeRef] = 0; #Above see level
            gps[piexif.GPSIFD.GPSAltitude]= (int(abs(altitude)*100),100)
  
        #Latitude
        if lat<0:
            gps[piexif.GPSIFD.GPSLatitudeRef] = "S"
        else:
            gps[piexif.GPSIFD.GPSLatitudeRef] = "N"
   
        latd,latm,lats = ExifWriter._decdeg2dms(lat)
        gps[piexif.GPSIFD.GPSLatitude] = [(latd,1),(latm,1),(lats*100,100)];
  
        #Longitude
        if longit<0:
            gps[piexif.GPSIFD.GPSLongitudeRef] = "W"
        else:
            gps[piexif.GPSIFD.GPSLongitudeRef] = "E"
  
        longd,longm,longs = ExifWriter._decdeg2dms(longit)
        gps[piexif.GPSIFD.GPSLongitude] = [(longd,1),(longm,1),(longs*100,100)];

        exifdict =  piexif.load(filename)
        exifdict["GPS"] = gps

        exif_bytes = piexif.dump(exifdict)
        piexif.insert(exif_bytes, filename)
        exifdict = piexif.load(filename)

#****************************************************************************
#   Method Name     : _decdeg2dms
#
#   Description     : Converts decimal degrees to Degrees, Minutes, seconds
#
#   Parameters      : dd    Coordinate in decimal degrees
#
#   Return Value    : None
#
#   Author           : Gael Billon, Escadrone
#
#****************************************************************************

    #method converting decimal format to DMS
    @staticmethod
    def _decdeg2dms(dd):
        dd = abs(dd)
        minutes,seconds = divmod(dd*3600,60)
        degrees,minutes = divmod(minutes,60)
        return (degrees,minutes,seconds)

 



'''Joystick control classes'''


def scale(val,
          inlow=-1, inhigh=1,
          outlow=1000, outhigh=2000):
    '''Scale an in value in the range (inlow, inhigh) to the
    range (outlow, outhigh).'''
    return (
        ((float(val) - inlow) / (inhigh - inlow)) *
        (outhigh - outlow) + outlow
    )


class Control (object):
    '''Base class for all controls'''
    def __init__(self, joystick,
                 inlow=-1, inhigh=1,
                 outlow=1000, outhigh=2000):
        self.joystick = joystick
        self.inlow = inlow
        self.inhigh = inhigh
        self.outlow = outlow
        self.outhigh = outhigh


class Button (Control):
    '''A Button acts like a momentary switch.  When pressed, the
    corresponding channel value is set to `outhigh`; when released,
    the value is set to `outlow`.'''

    def __init__(self, joystick, id, **kwargs):
        super(Button, self).__init__(joystick, **kwargs)
        self.id = id

    @property
    def value(self):
        state = self.joystick.get_button(self.id)
        if state:
            return self.outhigh
        else:
            return self.outlow


class MultiButton (Control):
    '''A MultiButton maps multiple buttons to the same channel like a
    multiple-position switch.  When a button is pressed, the channel
    is set to the corresponding value.  When a button is released, no
    changes is made to the channel.'''

    def __init__(self, joystick, buttons, **kwargs):
        super(MultiButton, self).__init__(joystick, **kwargs)
        self.buttons = buttons
        self._value = buttons[0]['value']

    @property
    def value(self):
        for button in self.buttons:
            state = self.joystick.get_button(button['id'])
            if state:
                self._value = button['value']
                break

        return self._value


class Axis (Control):
    '''An Axis maps a joystick axis to a channel.  Set `invert` to
    `True` in order to reverse the direction of the input.'''

    def __init__(self, joystick, id, invert=False, **kwargs):
        super(Axis, self).__init__(joystick, **kwargs)
        self.id = id
        self.invert = invert

    @property
    def value(self):
        val = self.joystick.get_axis(self.id)
        if self.invert:
            val = -val

        return scale(val, outlow=self.outlow, outhigh=self.outhigh)


class Hat (Control):
    '''A Hat maps one axis of a hat as if it were a toggle switch.
    When the axis goes negative, the corresponding channel value is
    set to `outputlow`.  When the axis goes positive, the value is set
    to `outputhigh`.  No change is made when the axis returns to 0.'''

    def __init__(self, joystick, id, axis, **kwargs):
        super(Hat, self).__init__(joystick, **kwargs)
        self.id = id
        self.axis = axis
        self._value = self.outlow

    @property
    def value(self):
        x, y = self.joystick.get_hat(self.id)

        value = x if self.axis == 'x' else y

        if value != 0:
            self._value = scale(value,
                                outlow=self.outlow, outhigh=self.outhigh)

        return self._value


class Joystick (object):
    '''A Joystick manages a collection of Controls.'''

    def __init__(self, joystick, controls):
        self.joystick = joystick
        self.controls = controls

        self.chan_max = max(control['channel']
                            for control in controls['controls'])
        self.channels = [None] * self.chan_max

        self.joystick.init()

        for control in controls['controls']:
            if control['type'] == 'button':
                kwargs = {k: control[k]
                          for k in control.keys()
                          if k in ['outlow', 'outhigh']}

                handler = Button(self.joystick, control['id'], **kwargs)

            elif control['type'] == 'axis':
                kwargs = {k: control[k]
                          for k in control.keys()
                          if k in ['inlow', 'inhigh',
                                   'outlow', 'outhigh', 'invert']}

                handler = Axis(self.joystick, control['id'], **kwargs)

            elif control['type'] == 'multibutton':
                handler = MultiButton(self.joystick,
                                   buttons=control['buttons'])
            elif control['type'] == 'hat':
                kwargs = {k: control[k]
                          for k in control.keys()
                          if k in ['outlow', 'outhigh']}

                handler = Hat(self.joystick, control['id'], control['axis'])

            self.channels[control['channel']-1] = handler

    def read(self):
        '''Returns an array of channel values.  Return 0 for channels
        not specified in the control definition.'''

        return [int(handler.value) if handler is not None else 0
                for handler in self.channels]



#!/usr/bin/env python
'''enable run-time addition and removal of UDP clients , just like --out on the cnd line'''
''' TO USE:
    output add 10.11.12.13:14550
    output list
    output remove 3      # to remove 3rd output
'''

from pymavlink import mavutil


from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util

class OutputModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(OutputModule, self).__init__(mpstate, "output", "output control", public=True)
        self.add_command('output', self.cmd_output, "output control",
                         ["<list|add|remove|sysid>"])

    def cmd_output(self, args):
        '''handle output commands'''
        if len(args) < 1 or args[0] == "list":
            self.cmd_output_list()
        elif args[0] == "add":
            if len(args) != 2:
                print("Usage: output add OUTPUT")
                return
            self.cmd_output_add(args[1:])
        elif args[0] == "remove":
            if len(args) != 2:
                print("Usage: output remove OUTPUT")
                return
            self.cmd_output_remove(args[1:])
        elif args[0] == "sysid":
            if len(args) != 3:
                print("Usage: output sysid SYSID OUTPUT")
                return
            self.cmd_output_sysid(args[1:])
        else:
            print("usage: output <list|add|remove|sysid>")

    def cmd_output_list(self):
        '''list outputs'''
        print("%u outputs" % len(self.mpstate.mav_outputs))
        for i in range(len(self.mpstate.mav_outputs)):
            conn = self.mpstate.mav_outputs[i]
            print("%u: %s" % (i, conn.address))
        if len(self.mpstate.sysid_outputs) > 0:
            print("%u sysid outputs" % len(self.mpstate.sysid_outputs))
            for sysid in self.mpstate.sysid_outputs:
                conn = self.mpstate.sysid_outputs[sysid]
                print("%u: %s" % (sysid, conn.address))

    def cmd_output_add(self, args):
        '''add new output'''
        device = args[0]
        print("Adding output %s" % device)
        try:
            conn = mavutil.mavlink_connection(device, input=False, source_system=self.settings.source_system)
            conn.mav.srcComponent = self.settings.source_component
        except Exception:
            print("Failed to connect to %s" % device)
            return
        self.mpstate.mav_outputs.append(conn)
        try:
            mp_util.child_fd_list_add(conn.port.fileno())
        except Exception:
            pass

    def cmd_output_sysid(self, args):
        '''add new output for a specific MAVLink sysID'''
        sysid = int(args[0])
        device = args[1]
        print("Adding output %s for sysid %u" % (device, sysid))
        try:
            conn = mavutil.mavlink_connection(device, input=False, source_system=self.settings.source_system)
            conn.mav.srcComponent = self.settings.source_component
        except Exception:
            print("Failed to connect to %s" % device)
            return
        try:
            mp_util.child_fd_list_add(conn.port.fileno())
        except Exception:
            pass
        if sysid in self.mpstate.sysid_outputs:
            self.mpstate.sysid_outputs[sysid].close()
        self.mpstate.sysid_outputs[sysid] = conn

    def cmd_output_remove(self, args):
        '''remove an output'''
        device = args[0]
        for i in range(len(self.mpstate.mav_outputs)):
            conn = self.mpstate.mav_outputs[i]
            if str(i) == device or conn.address == device:
                print("Removing output %s" % conn.address)
                try:
                    mp_util.child_fd_list_add(conn.port.fileno())
                except Exception:
                    pass
                conn.close()
                self.mpstate.mav_outputs.pop(i)
                return

    def idle_task(self):
        '''called on idle'''
        for m in self.mpstate.mav_outputs:
            m.source_system = self.settings.source_system
            m.mav.srcSystem = m.source_system
            m.mav.srcComponent = self.settings.source_component

def init(mpstate):
    '''initialise module'''
    return OutputModule(mpstate)



#!/usr/bin/env python

'''
view a mission log on a map
'''

import sys, time, os
import re

from math import *

from pymavlink import mavutil, mavwp, mavextra
from MAVProxy.modules.mavproxy_map import mp_slipmap, mp_tile
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.lib import grapher
from MAVProxy.modules.lib import kmlread
import functools
import random
import cv2

def create_map(title):
    '''create map object'''

def pixel_coords(latlon, ground_width=0, mt=None, topleft=None, width=None):
    '''return pixel coordinates in the map image for a (lat,lon)'''
    (lat,lon) = (latlon[0], latlon[1])
    return mt.coord_to_pixel(topleft[0], topleft[1], width, ground_width, lat, lon)

def create_imagefile(options, filename, latlon, ground_width, path_objs, mission_obj, fence_obj, kml_objects, width=600, height=600, used_flightmodes=[], mav_type=None):
    '''create path and mission as an image file'''
    mt = mp_tile.MPTile(service=options.service)

    map_img = mt.area_to_image(latlon[0], latlon[1],
                               width, height, ground_width)
    while mt.tiles_pending() > 0:
        print("Waiting on %u tiles" % mt.tiles_pending())
        time.sleep(1)
    map_img = mt.area_to_image(latlon[0], latlon[1],
                               width, height, ground_width)
    # a function to convert from (lat,lon) to (px,py) on the map
    pixmapper = functools.partial(pixel_coords, ground_width=ground_width, mt=mt, topleft=latlon, width=width)
    for path_obj in path_objs:
        path_obj.draw(map_img, pixmapper, None)
    if mission_obj is not None:
        for m in mission_obj:
            m.draw(map_img, pixmapper, None)
    if fence_obj is not None:
        fence_obj.draw(map_img, pixmapper, None)

    if kml_objects is not None:
        for obj in kml_objects:
            print(obj)
            try:
                obj.draw(map_img, pixmapper, None)
            except Exception:
                pass

    if (options is not None and
        mav_type is not None and
        options.colour_source == "flightmode"):
        tuples = [ (mode, colour_for_flightmode(mav_type, mode))
                   for mode in used_flightmodes.keys() ]
        legend = mp_slipmap.SlipFlightModeLegend("legend", tuples)
        legend.draw(map_img, pixmapper, None)

    map_img = cv2.cvtColor(map_img, cv2.COLOR_BGR2RGB)
    cv2.imwrite(filename, map_img)

map_colours = [ (255,   0,   0),
                (  0, 255,   0),
                (  0,   0, 255),
                (127,   0,   0),
                (  0, 127,   0),
#                (  0,   0, 127),
                (255, 255,   0),
                (255,   0, 255),
                (  0, 255, 255),
                (127, 127,   0),
                (127,   0, 127),
                (  0, 127, 127),

                (255,  191,   0),
                (255,   0,  191),
                (255,  191,  191),

                (191, 255,   0),
                (191,   0, 255),
                (191, 255, 255),

                ( 0, 255,  191),
                ( 0,  191, 255),
                ( 0, 255, 255),
]
colour_map_copter = {}
colour_map_plane = {}
colour_map_rover = {}
colour_map_tracker = {}
colour_map_submarine = {}
colour_map_blimp = {}

for mytuple in ((mavutil.mode_mapping_apm.values(),colour_map_plane),
                (mavutil.mode_mapping_acm.values(),colour_map_copter),
                (mavutil.mode_mapping_rover.values(),colour_map_rover),
                (mavutil.mode_mapping_tracker.values(),colour_map_tracker),
                (mavutil.mode_mapping_sub.values(),colour_map_submarine),
                (mavutil.mode_mapping_blimp.values(),colour_map_blimp),
):
    (mode_names, colour_map) = mytuple
    i=0
    for mode_name in mode_names:
        colour_map[mode_name] = map_colours[i]
        i += 1
        if i >= len(map_colours):
            print("Warning: reusing colours!")
            i = 0
    colour_map["UNKNOWN"] = (0, 0, 0)

colourmap_check_done = False
def colourmap_for_mav_type(mav_type):
    # swiped from "def mode_mapping_byname(mav_type):" in mavutil
    map = None
    if mav_type in [mavutil.mavlink.MAV_TYPE_QUADROTOR,
                    mavutil.mavlink.MAV_TYPE_HELICOPTER,
                    mavutil.mavlink.MAV_TYPE_HEXAROTOR,
                    mavutil.mavlink.MAV_TYPE_OCTOROTOR,
                    mavutil.mavlink.MAV_TYPE_COAXIAL,
                    mavutil.mavlink.MAV_TYPE_TRICOPTER]:
        map = colour_map_copter
    if mav_type == mavutil.mavlink.MAV_TYPE_FIXED_WING:
        map = colour_map_plane
    if mav_type in [mavutil.mavlink.MAV_TYPE_GROUND_ROVER,
                    mavutil.mavlink.MAV_TYPE_SURFACE_BOAT]:
        map = colour_map_rover
    if mav_type == mavutil.mavlink.MAV_TYPE_ANTENNA_TRACKER:
        map = colour_map_tracker
    if mav_type == mavutil.mavlink.MAV_TYPE_SUBMARINE:
        map = colour_map_submarine
    if mav_type == mavutil.mavlink.MAV_TYPE_AIRSHIP:
        map = colour_map_blimp
    if map is None:
        print("No colormap for mav_type=%u" % (mav_type,))
        # we probably don't have a valid mode map, so returning
        # anything but the empty map here is probably pointless:
        map = colour_map_plane
    return map

def display_waypoints(wploader, map):
    '''display the waypoints'''
    mission_list = wploader.view_list()
    polygons = wploader.polygon_list()
    map.add_object(mp_slipmap.SlipClearLayer('Mission'))
    for k in range(len(polygons)):
        p = polygons[k]
        if len(p) > 1:
            map.add_object(mp_slipmap.SlipPolygon('mission %u' % k, p,
                                                  layer='Mission', linewidth=2, colour=(255,255,255)))
        labeled_wps = {}
        for i in range(len(mission_list)):
            next_list = mission_list[i]
            for j in range(len(next_list)):
                #label already printed for this wp?
                if (next_list[j] not in labeled_wps):
                    map.add_object(mp_slipmap.SlipLabel(
                        'miss_cmd %u/%u' % (i,j), polygons[i][j], str(next_list[j]), 'Mission', colour=(0,255,255)))
                    labeled_wps[next_list[j]] = (i,j)

colour_expression_exceptions = dict()
colour_source_min = 255
colour_source_max = 0

def colour_for_point(mlog, point, instance, options):
    global colour_expression_exceptions, colour_source_max, colour_source_min
    '''indicate a colour to be used to plot point'''
    source = getattr(options, "colour_source", "flightmode")
    if source == "flightmode":
        return colour_for_point_flightmode(mlog, point, instance, options)
    elif source == "type":
        return colour_for_point_type(mlog, point, instance, options)

    # evaluate source as an expression which should return a
    # number in the range 0..255
    try:
        v = eval(source, globals(),mlog.messages)
    except Exception as e:
        str_e = str(e)
        try:
            count = colour_expression_exceptions[str_e]
        except KeyError:
            colour_expression_exceptions[str_e] = 0
            count = 0
        if count > 100:
            print("Too many exceptions processing (%s): %s" % (source, str_e))
            sys.exit(1)
        colour_expression_exceptions[str_e] += 1
        v = 0

    # we don't use evaluate_expression as we want the exceptions...
#    v = mavutil.evaluate_expression(source, mlog.messages)

    if v is None:
        return v
    if isinstance(v, str):
        print("colour expression returned a string: %s" % v)
        sys.exit(1)
    elif v < 0:
        print("colour expression returned %d (< 0)" % v)
        v = 0
    elif v > 255:
        print("colour expression returned %d (> 255)" % v)
        v = 255

    if v < colour_source_min:
        colour_source_min = v
    if v > colour_source_max:
        colour_source_max = v

    r = 255
    g = 255
    b = v
    return (b,b,b)

def colour_for_point_flightmode(mlog, point, instance, options):
    return colour_for_flightmode(getattr(mlog, 'mav_type',None), getattr(mlog, 'flightmode',''), instance)

def colour_for_flightmode(mav_type, fmode, instance=0):
    colourmap = colourmap_for_mav_type(mav_type)
    if fmode in colourmap:
        colour = colourmap[fmode]
    else:
        print("No entry in colourmap for %s" % (str(fmode)))
        colour = colourmap['UNKNOWN']
    (r,g,b) = colour
    (r,g,b) = (r+instance*80,g+instance*50,b+instance*70)
    if r > 255:
        r = 205
    if g > 255:
        g = 205
    if g < 0:
        g = 0
    if b > 255:
        b = 205
    colour = (r,g,b)
    return colour

def colour_for_point_type(mlog, point, instance, options):
    colors=[ 'red', 'green', 'blue', 'orange', 'olive', 'cyan', 'magenta', 'brown',
             'violet', 'purple', 'grey', 'black']
    return map_colours[instance]

def mavflightview_mav(mlog, options=None, flightmode_selections=[]):
    '''create a map for a log file'''
    wp = mavwp.MAVWPLoader()
    if options.mission is not None:
        wp.load(options.mission)
    fen = mavwp.MAVFenceLoader()
    if options.fence is not None:
        fen.load(options.fence)
    all_false = True
    for s in flightmode_selections:
        if s:
            all_false = False
    idx = 0
    path = [[]]
    instances = {}
    ekf_counter = 0
    nkf_counter = 0
    types = ['MISSION_ITEM', 'MISSION_ITEM_INT', 'CMD']
    if options.types is not None:
        types.extend(options.types.split(','))
    else:
        types.extend(['POS','GLOBAL_POSITION_INT'])
        if options.rawgps or options.dualgps:
            types.extend(['GPS', 'GPS_RAW_INT'])
        if options.rawgps2 or options.dualgps:
            types.extend(['GPS2_RAW','GPS2'])
        if options.ekf:
            types.extend(['EKF1', 'GPS'])
        if options.nkf:
            types.extend(['NKF1', 'GPS'])
        if options.ahr2:
            types.extend(['AHR2', 'AHRS2', 'GPS'])

    # handle forms like GPS[0], mapping to GPS for recv_match_types
    for i in range(len(types)):
        bracket = types[i].find('[')
        if bracket != -1:
            types[i] = types[i][:bracket]

    recv_match_types = types[:]
    colour_source = getattr(options, "colour_source")
    re_caps = re.compile('[A-Z_][A-Z0-9_]+')

    if colour_source is not None:
        # stolen from mavgraph.py
        caps = set(re.findall(re_caps, colour_source))
        recv_match_types.extend(caps)

    print("Looking for types %s" % str(recv_match_types))

    last_timestamps = {}
    used_flightmodes = {}

    mlog.rewind()

    while True:
        try:
            m = mlog.recv_match(type=recv_match_types)
            if m is None:
                break
        except Exception:
            break

        type = m.get_type()

        if type in ['MISSION_ITEM', 'MISSION_ITEM_INT']:
            try:
                new_m = m
                if type == 'MISSION_ITEM_INT':
                    # create a MISSION_ITEM from MISSION_ITEM_INT
                    new_m = mavutil.mavlink.MAVLink_mission_item_message(
                        0,
                        0,
                        m.seq,
                        m.frame,
                        m.command,
                        m.current,
                        m.autocontinue,
                        m.param1,
                        m.param2,
                        m.param3,
                        m.param4,
                        m.x / 1.0e7,
                        m.y / 1.0e7,
                        m.z
                    )
                while new_m.seq > wp.count():
                    print("Adding dummy WP %u" % wp.count())
                    wp.set(new_m, wp.count())
                wp.set(new_m, m.seq)
            except Exception as e:
                print("Exception: %s" % str(e))
                pass
            continue
        elif type == 'CMD':
            if options.mission is None:
                m = mavutil.mavlink.MAVLink_mission_item_message(0,
                                                                0,
                                                                m.CNum,
                                                                mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                                m.CId,
                                                                0, 1,
                                                                m.Prm1, m.Prm2, m.Prm3, m.Prm4,
                                                                m.Lat, m.Lng, m.Alt)
                try:
                    while m.seq > wp.count():
                        print("Adding dummy WP %u" % wp.count())
                        wp.set(m, wp.count())
                    wp.set(m, m.seq)
                except Exception:
                    pass
            continue
        if not mlog.check_condition(options.condition):
            continue
        if options.mode is not None and mlog.flightmode.lower() != options.mode.lower():
            continue

        if not type in types:
            # may only be present for colour-source expressions to work
            continue

        if type == 'GPS' and hasattr(m,'I'):
            type = 'GPS[%u]' % m.I

        if not all_false and len(flightmode_selections) > 0 and idx < len(options._flightmodes) and m._timestamp >= options._flightmodes[idx][2]:
            idx += 1
        elif (idx < len(flightmode_selections) and flightmode_selections[idx]) or all_false or len(flightmode_selections) == 0:
            used_flightmodes[mlog.flightmode] = 1
            (lat, lng) = (None,None)
            if type in ['GPS','GPS2']:
                status = getattr(m, 'Status', None)
                nsats = getattr(m, 'NSats', None)
                if status is None:
                    status = getattr(m, 'FixType', None)
                    if status is None:
                        print("Can't find status on GPS message")
                        print(m)
                        break
                if status < 2 and nsats < 5:
                    continue
                # flash log
                lat = m.Lat
                lng = getattr(m, 'Lng', None)
                if lng is None:
                    lng = getattr(m, 'Lon', None)
                    if lng is None:
                        print("Can't find longitude on GPS message")
                        print(m)
                        break
            elif type in ['EKF1', 'ANU1']:
                pos = mavextra.ekf1_pos(m)
                if pos is None:
                    continue
                ekf_counter += 1
                if ekf_counter % options.ekf_sample != 0:
                    continue
                (lat, lng) = pos
            elif type in ['NKF1','XKF1']:
                pos = mavextra.ekf1_pos(m)
                if pos is None:
                    continue
                nkf_counter += 1
                if nkf_counter % options.nkf_sample != 0:
                    continue
                (lat, lng) = pos
            elif type in ['ANU5']:
                (lat, lng) = (m.Alat*1.0e-7, m.Alng*1.0e-7)
            elif type in ['AHR2', 'POS', 'CHEK']:
                (lat, lng) = (m.Lat, m.Lng)
            elif type == 'AHRS2':
                (lat, lng) = (m.lat*1.0e-7, m.lng*1.0e-7)
            elif type == 'ORGN':
                (lat, lng) = (m.Lat, m.Lng)
            elif type == 'SIM':
                (lat, lng) = (m.Lat, m.Lng)
            elif type == 'GUID':
                if (m.Type == 0):
                    (lat, lng) = (m.pX*1.0e-7, m.pY*1.0e-7)
            else:
                if hasattr(m,'Lat'):
                    lat = m.Lat
                if hasattr(m,'Lon'):
                    lng = m.Lon
                if hasattr(m,'Lng'):
                    lng = m.Lng
                if hasattr(m,'lat'):
                    lat = m.lat * 1.0e-7
                if hasattr(m,'lon'):
                    lng = m.lon * 1.0e-7
                if hasattr(m,'latitude'):
                    lat = m.latitude * 1.0e-7
                if hasattr(m,'longitude'):
                    lng = m.longitude * 1.0e-7

            if lat is None or lng is None:
                continue

            # automatically add new types to instances
            if type not in instances:
                instances[type] = len(instances)
                while len(instances) >= len(path):
                    path.append([])
            instance = instances[type]

            # only plot thing we have a valid-looking location for:
            if abs(lat)<=0.01 and abs(lng)<=0.01:
                continue

            colour = colour_for_point(mlog, (lat, lng), instance, options)
            if colour is None:
                continue

            tdays = grapher.timestamp_to_days(m._timestamp)
            point = (lat, lng, colour, tdays)

            if options.rate == 0 or not type in last_timestamps or m._timestamp - last_timestamps[type] > 1.0/options.rate:
                last_timestamps[type] = m._timestamp
                path[instance].append(point)
    if len(path[0]) == 0:
        print("No points to plot")
        return None

    return [path, wp, fen, used_flightmodes, getattr(mlog, 'mav_type',None), instances]

def mavflightview_show(path, wp, fen, used_flightmodes, mav_type, options, instances, title=None, timelim_pipe=None, show_waypoints=True):
    if not title:
        title='MAVFlightView'


    boundary_path = []
    for p in path[0]:
        boundary_path.append((p[0],p[1]))

    fence = fen.polygon()
    if options.fencebounds:
        for p in fence:
            boundary_path.append((p[0],p[1]))

    bounds = mp_util.polygon_bounds(boundary_path)
    (lat, lon) = (bounds[0]+bounds[2], bounds[1])
    (lat, lon) = mp_util.gps_newpos(lat, lon, -45, 50)
    ground_width = mp_util.gps_distance(lat, lon, lat-bounds[2], lon+bounds[3])
    while (mp_util.gps_distance(lat, lon, bounds[0], bounds[1]) >= ground_width-20 or
           mp_util.gps_distance(lat, lon, lat, bounds[1]+bounds[3]) >= ground_width-20):
        ground_width += 10

    path_objs = []
    for i in range(len(path)):
        if len(path[i]) != 0:
            path_objs.append(mp_slipmap.SlipPolygon('FlightPath[%u]-%s' % (i,title), path[i], layer='FlightPath',
                                                    linewidth=2, colour=(255,0,180)))
    plist = []
    if options.show_waypoints:
        plist = wp.polygon_list()
    mission_obj = None
    if len(plist) > 0:
        mission_obj = []
        for i in range(len(plist)):
            mission_obj.append(mp_slipmap.SlipPolygon('Mission-%s-%u' % (title,i), plist[i], layer='Mission',
                                                      linewidth=2, colour=(255,255,255)))
    else:
        mission_obj = None

    if len(fence) > 1:
        fence_obj = mp_slipmap.SlipPolygon('Fence-%s' % title, fen.polygon(), layer='Fence',
                                           linewidth=2, colour=(0,255,0))
    else:
        fence_obj = None

    kml = getattr(options,'kml',None)
    if kml is not None:
        kml_objects = load_kml(kml)
    else:
        kml_objects = None

    if options.imagefile:
        create_imagefile(options, options.imagefile, (lat,lon), ground_width, path_objs, mission_obj, fence_obj, kml_objects, used_flightmodes=used_flightmodes, mav_type=mav_type)
    else:
        global multi_map
        if options.multi and multi_map is not None:
            map = multi_map
        else:
            map = mp_slipmap.MPSlipMap(title=title,
                                       service=options.service,
                                       elevation=True,
                                       width=600,
                                       height=600,
                                       ground_width=ground_width,
                                       lat=lat, lon=lon,
                                       debug=options.debug,
                                       show_flightmode_legend=options.show_flightmode_legend,
                                       timelim_pipe=timelim_pipe)
        if options.multi:
            multi_map = map
        for path_obj in path_objs:
            map.add_object(path_obj)
        if mission_obj is not None:
            display_waypoints(wp, map)
        if fence_obj is not None:
            map.add_object(fence_obj)

        if kml_objects is not None:
            for obj in kml_objects:
                map.add_object(obj)
            
        for flag in options.flag:
            a = flag.split(',')
            lat = a[0]
            lon = a[1]
            icon = 'flag.png'
            if len(a) > 2:
                icon = a[2] + '.png'
            icon = map.icon(icon)
            map.add_object(mp_slipmap.SlipIcon('icon - %s' % str(flag), (float(lat),float(lon)), icon, layer=3, rotation=0, follow=False))

        if options.colour_source == "flightmode":
            tuples = [ (mode, colour_for_flightmode(mav_type, mode))
                       for mode in used_flightmodes.keys() ]
            map.add_object(mp_slipmap.SlipFlightModeLegend("legend", tuples))
        elif options.colour_source == "type":
            tuples = [ (t, map_colours[instances[t]]) for t in instances.keys() ]
            map.add_object(mp_slipmap.SlipFlightModeLegend("legend", tuples))
        else:
            print("colour-source: min=%f max=%f" % (colour_source_min, colour_source_max))


def load_kml(kml):
    '''load a kml overlay, return list of map objects'''
    print("Loading kml %s" % kml)
    nodes = kmlread.readkmz(kml)
    ret = []
    for n in nodes:
        try:
            point = kmlread.readObject(n)
        except Exception as ex:
            continue

        if point[0] == 'Polygon':
            newcolour = (random.randint(0, 255), 0, random.randint(0, 255))
            curpoly = mp_slipmap.SlipPolygon(point[1], point[2],
                                             layer=2, linewidth=2, colour=newcolour)
            ret.append(curpoly)

        if point[0] == 'Point':
            icon = mp_tile.mp_icon('barrell.png')
            curpoint = mp_slipmap.SlipIcon(point[1], latlon = (point[2][0][0], point[2][0][1]), layer=3, img=icon, rotation=0, follow=False)
            curtext = mp_slipmap.SlipLabel(point[1], point = (point[2][0][0], point[2][0][1]), layer=4, label=point[1], colour=(0,255,255))
            ret.append(curpoint)
            ret.append(curtext)
    return ret


def mavflightview(filename, options):
    print("Loading %s ..." % filename)
    mlog = mavutil.mavlink_connection(filename)
    stuff = mavflightview_mav(mlog, options)
    if stuff is None:
        return
    [path, wp, fen, used_flightmodes, mav_type, instances] = stuff
    mavflightview_show(path, wp, fen, used_flightmodes, mav_type, options, instances, title=filename)

class mavflightview_options(object):
    def __init__(self):
        self.service = "MicrosoftHyb"
        self.mode = None
        self.condition = None
        self.mission = None
        self.fence = None
        self.fencebounds = False
        self.imagefile = None
        self.flag = []
        self.rawgps = False
        self.rawgps2 = False
        self.dualgps = False
        self.ekf = False
        self.nkf = False
        self.ahr2 = False
        self.debug = False
        self.multi = False
        self.types = None
        self.ekf_sample = 1
        self.rate = 0
        self._flightmodes = []
        self.colour_source = 'flightmode'
        self.show_waypoints = True

if __name__ == "__main__":
    multiproc.freeze_support()

    from optparse import OptionParser
    parser = OptionParser("mavflightview.py [options]")
    parser.add_option("--service", default="MicrosoftSat", help="tile service")
    parser.add_option("--mode", default=None, help="flight mode")
    parser.add_option("--condition", default=None, help="conditional check on log")
    parser.add_option("--mission", default=None, help="mission file (defaults to logged mission)")
    parser.add_option("--fence", default=None, help="fence file")
    parser.add_option("--fencebounds", action='store_true', help="use fence boundary for zoom")
    parser.add_option("--imagefile", default=None, help="output to image file")
    parser.add_option("--flag", default=[], type='str', action='append', help="flag positions")
    parser.add_option("--rawgps", action='store_true', default=False, help="use GPS_RAW_INT")
    parser.add_option("--rawgps2", action='store_true', default=False, help="use GPS2_RAW")
    parser.add_option("--dualgps", action='store_true', default=False, help="use GPS_RAW_INT and GPS2_RAW")
    parser.add_option("--ekf", action='store_true', default=False, help="use EKF1 pos")
    parser.add_option("--nkf", action='store_true', default=False, help="use NKF1 pos")
    parser.add_option("--ahr2", action='store_true', default=False, help="use AHR2 pos")
    parser.add_option("--debug", action='store_true', default=False, help="show debug info")
    parser.add_option("--multi", action='store_true', default=False, help="show multiple flights on one map")
    parser.add_option("--types", default=None, help="types of position messages to show")
    parser.add_option("--ekf-sample", type='int', default=1, help="sub-sampling of EKF messages")
    parser.add_option("--nkf-sample", type='int', default=1, help="sub-sampling of NKF messages")
    parser.add_option("--rate", type='int', default=0, help="maximum message rate to display (0 means all points)")
    parser.add_option("--colour-source", type="str", default="flightmode", help="expression with range 0f..255f used for point colour")
    parser.add_option("--no-flightmode-legend", action="store_false", default=True, dest="show_flightmode_legend", help="hide legend for colour used for flight modes")
    parser.add_option("--kml", default=None, help="add kml overlay")
    parser.add_option("--hide-waypoints", dest='show_waypoints', action='store_false', help="do not show waypoints", default=True)

    (opts, args) = parser.parse_args()

    try:
        import faulthandler, signal
        try:
            faulthandler.register(signal.SIGUSR1)
        except AttributeError as e:
            pass
    except ImportError:
        pass

    if len(args) < 1:
        print("Usage: mavflightview.py [options] <LOGFILE...>")
        sys.exit(1)

    if opts.multi:
        multi_map = None

    random.seed(1)

    for f in args:
        mavflightview(f, opts)



import errno
import time
import os
import socket
from MAVProxy.modules.lib import mp_menu
from MAVProxy.modules.lib.wxconsole_util import Value, Text
from MAVProxy.modules.lib.wx_loader import wx
from MAVProxy.modules.lib import win_layout
from MAVProxy.modules.lib import icon

class ConsoleFrame(wx.Frame):
    """ The main frame of the console"""

    def __init__(self, state, title):
        self.state = state
        wx.Frame.__init__(self, None, title=title, size=(800,300))
        # different icons for MAVExplorer and MAVProxy
        try:
            if title == "MAVExplorer":
                self.SetIcon(icon.SimpleIcon("EXPLORER").get_ico())
            else:
                self.SetIcon(icon.SimpleIcon("CONSOLE").get_ico())
        except Exception:
            pass
        self.panel = wx.Panel(self)
        self.panel.SetBackgroundColour('white')
        state.frame = self

        # values for the status bar
        self.values = {}

        self.menu = None
        self.menu_callback = None
        self.last_layout_send = time.time()

        self.control = wx.TextCtrl(self.panel, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_AUTO_URL)

        self.vbox = wx.BoxSizer(wx.VERTICAL)
        # start with one status row
        self.status = [wx.BoxSizer(wx.HORIZONTAL)]
        self.vbox.Add(self.status[0], 0, flag=wx.ALIGN_LEFT | wx.TOP)
        self.vbox.Add(self.control, 1, flag=wx.LEFT | wx.BOTTOM | wx.GROW)

        self.panel.SetSizer(self.vbox)

        self.timer = wx.Timer(self)

        self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
        self.timer.Start(100)

        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.Bind(wx.EVT_TEXT_URL, self.on_text_url)

        self.Show(True)
        self.pending = []

    def on_menu(self, event):
        '''handle menu selections'''
        state = self.state
        ret = self.menu.find_selected(event)
        if ret is None:
            return
        ret.call_handler()
        state.child_pipe_send.send(ret)

    def on_text_url(self, event):
        '''handle double clicks on URL text'''
        try:
            import webbrowser
        except ImportError:
            return
        mouse_event = event.GetMouseEvent()
        if mouse_event.LeftDClick():
            url_start = event.GetURLStart()
            url_end = event.GetURLEnd()
            url = self.control.GetRange(url_start, url_end)
            try:
                # attempt to use google-chrome
                browser_controller = webbrowser.get('google-chrome')
                browser_controller.open_new_tab(url)
            except webbrowser.Error:
                # use the system configured default browser
                webbrowser.open_new_tab(url)

    def on_idle(self, event):
        time.sleep(0.05)
        now = time.time()
        if now - self.last_layout_send > 1:
            self.last_layout_send = now
            self.state.child_pipe_send.send(win_layout.get_wx_window_layout(self))

    def on_timer(self, event):
        state = self.state
        if state.close_event.wait(0.001):
            self.timer.Stop()
            self.Destroy()
            return

        while True:
            try:
                poll_success = state.child_pipe_recv.poll()
                if not poll_success:
                    return
            except socket.error as e:
                if e.errno == errno.EPIPE:
                    break
                else:
                    raise e

            try:
                obj = state.child_pipe_recv.recv()
            except EOFError:
                break
                
            if isinstance(obj, Value):
                # request to set a status field
                if not obj.name in self.values:
                    # create a new status field
                    value = wx.StaticText(self.panel, -1, obj.text)
                    # possibly add more status rows
                    for i in range(len(self.status), obj.row+1):
                        self.status.append(wx.BoxSizer(wx.HORIZONTAL))
                        self.vbox.Insert(len(self.status)-1, self.status[i], 0, flag=wx.ALIGN_LEFT | wx.TOP)
                        self.vbox.Layout()
                    self.status[obj.row].Add(value, border=5)
                    self.status[obj.row].AddSpacer(20)
                    self.values[obj.name] = value
                value = self.values[obj.name]
                value.SetForegroundColour(obj.fg)
                value.SetBackgroundColour(obj.bg)
                value.SetLabel(obj.text)
                self.panel.Layout()
            elif isinstance(obj, Text):
                '''request to add text to the console'''
                self.pending.append(obj)
                for p in self.pending:
                    # we're scrolled at the bottom
                    oldstyle = self.control.GetDefaultStyle()
                    style = wx.TextAttr()
                    style.SetTextColour(p.fg)
                    style.SetBackgroundColour(p.bg)
                    self.control.SetDefaultStyle(style)
                    self.control.AppendText(p.text)
                    self.control.SetDefaultStyle(oldstyle)
                self.pending = []
            elif isinstance(obj, mp_menu.MPMenuTop):
                if obj is not None:
                    self.SetMenuBar(None)
                    self.menu = obj
                    self.SetMenuBar(self.menu.wx_menu())
                    self.Bind(wx.EVT_MENU, self.on_menu)
                self.Refresh()
                self.Update()
            elif isinstance(obj, win_layout.WinLayout):
                win_layout.set_wx_window_layout(self, obj)

        self.timer.Stop()
        self.Destroy()
        return



#!/usr/bin/env python
'''rc command handling'''

import time, os, struct
from pymavlink import mavutil
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

class RCModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(RCModule, self).__init__(mpstate, "rc", "rc command handling", public = True)
        self.count = 18
        self.override = [ 0 ] * self.count
        self.last_override = [ 0 ] * self.count
        self.override_counter = 0
        x = "|".join(str(x) for x in range(1, (self.count+1)))
        self.add_command('rc', self.cmd_rc, "RC input control", ['<%s|all>' % x])
        self.add_command('switch', self.cmd_switch, "flight mode switch control", ['<0|1|2|3|4|5|6>'])
        self.rc_settings = mp_settings.MPSettings(
            [('override_hz', float, 10.0)])
        if self.sitl_output:
            self.rc_settings.override_hz = 20.0
        self.add_completion_function('(RCSETTING)',
                                     self.rc_settings.completion)
        self.override_period = mavutil.periodic_event(self.rc_settings.override_hz)

    def idle_task(self):
        self.override_period.frequency = self.rc_settings.override_hz
        if self.override_period.trigger():
            if (self.override != [ 0 ] * self.count or
                self.override != self.last_override or
                self.override_counter > 0):
                self.last_override = self.override[:]
                self.send_rc_override()
                if self.override_counter > 0:
                    self.override_counter -= 1

    def send_rc_override(self):
        '''send RC override packet'''
        if self.sitl_output:
            chan16 = self.override[:16]
            buf = struct.pack('<HHHHHHHHHHHHHHHH', *chan16)
            self.sitl_output.write(buf)
        else:
            chan18 = self.override[:18]
            self.master.mav.rc_channels_override_send(self.target_system,
                                                      self.target_component,
                                                      *chan18)

    def cmd_switch(self, args):
        '''handle RC switch changes'''
        mapping = [ 0, 1165, 1295, 1425, 1555, 1685, 1815 ]
        if len(args) != 1:
            print("Usage: switch <pwmvalue>")
            return
        value = int(args[0])
        if value < 0 or value > 6:
            print("Invalid switch value. Use 1-6 for flight modes, '0' to disable")
            return
        if self.vehicle_type == 'copter':
            default_channel = 5
        else:
            default_channel = 8
        if self.vehicle_type == 'rover':
            flite_mode_ch_parm = int(self.get_mav_param("MODE_CH", default_channel))
        else:
            flite_mode_ch_parm = int(self.get_mav_param("FLTMODE_CH", default_channel))
        self.override[flite_mode_ch_parm - 1] = mapping[value]
        self.override_counter = 10
        self.send_rc_override()
        if value == 0:
            print("Disabled RC switch override")
        else:
            print("Set RC switch override to %u (PWM=%u channel=%u)" % (
                value, mapping[value], flite_mode_ch_parm))

    def set_override(self, newchannels):
        '''this is a public method for use by drone API or other scripting'''
        self.override = newchannels
        self.override_counter = 10
        self.send_rc_override()

    def set_override_chan(self, channel, value):
        '''this is a public method for use by drone API or other scripting'''
        self.override[channel] = value
        self.override_counter = 10
        self.send_rc_override()

    def get_override_chan(self, channel):
        '''this is a public method for use by drone API or other scripting'''
        return self.override[channel]

    def cmd_rc(self, args):
        '''handle RC value override'''
        if len(args) > 0 and args[0] == 'set':
            self.rc_settings.command(args[1:])
            return
        if len(args) == 1 and args[0] == 'clear':
            channels = self.override
            for i in range(self.count):
                channels[i] = 0
            self.set_override(channels)
            return
        if len(args) != 2:
            print("Usage: rc <set|channel|all|clear> <pwmvalue>")
            return
        value = int(args[1])
        if value > 65535 or value < -1:
            raise ValueError("PWM value must be a positive integer between 0 and 65535")
        if value == -1:
            value = 65535
        channels = self.override
        if args[0] == 'all':
            for i in range(self.count):
                channels[i] = value
        else:
            channel = int(args[0])
            if channel < 1 or channel > self.count:
                print("Channel must be between 1 and %u or 'all'" % self.count)
                return
            channels[channel - 1] = value
        self.set_override(channels)

def init(mpstate):
    '''initialise module'''
    return RCModule(mpstate)



#!/usr/bin/env python
'''
camera view module
Malcolm Gill
Feb 2014
'''

import math
from MAVProxy.modules.mavproxy_map import mp_slipmap
from MAVProxy.modules.mavproxy_map import mp_elevation
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings
from cuav.lib import cuav_util
from cuav.camera.cam_params import CameraParams

# documented in common.xml, can't find these constants in code
scale_latlon = 1e-7
scale_hdg = 1e-2
scale_relative_alt = 1e-3

from MAVProxy.modules.lib import mp_module

class CameraViewModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(CameraViewModule, self).__init__(mpstate, "cameraview")
        self.add_command('cameraview', self.cmd_cameraview, "camera view")
        self.roll = 0
        self.pitch = 0
        self.yaw = 0
        self.mount_roll = 0
        self.mount_pitch = 0
        self.mount_yaw = 0
        self.height = 0
        self.lat = 0
        self.lon = 0
        self.home_height = 0
        self.hdg = 0
        self.elevation_model = mp_elevation.ElevationModel()
        self.camera_params = CameraParams() # TODO how to get actual camera params
        self.view_settings = mp_settings.MPSettings(
            [ ('r', float, 0.5),
              ('g', float, 0.5),
              ('b', float, 1.0),
            ])
        self.update_col()

    def update_col(self):
        self.col = tuple(int(255*c) for c in (self.view_settings.r, self.view_settings.g, self.view_settings.b))

    def cmd_cameraview(self, args):
        '''camera view commands'''
        state = self
        if args and args[0] == 'set':
            if len(args) < 3:
                state.view_settings.show_all()
            else:
                state.view_settings.set(args[1], args[2])
                state.update_col()
        else:
            print('usage: cameraview set')

    def unload(self):
        '''unload module'''
        pass

    def scale_rc(self, servo, min, max, param):
        '''scale a PWM value'''
        # default to servo range of 1000 to 2000
        min_pwm  = self.get_mav_param('%s_MIN'  % param, 0)
        max_pwm  = self.get_mav_param('%s_MAX'  % param, 0)
        if min_pwm == 0 or max_pwm == 0:
            return 0
        if max_pwm == min_pwm:
            p = 0.0
        else:
            p = (servo-min_pwm) / float(max_pwm-min_pwm)
        v = min + p*(max-min)
        if v < min:
            v = min
        if v > max:
            v = max
        return v

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        state = self
        mtype = m.get_type()
        if mtype == 'GLOBAL_POSITION_INT':
            state.lat, state.lon = m.lat*scale_latlon, m.lon*scale_latlon
            state.hdg = m.hdg*scale_hdg
            agl = state.elevation_model.GetElevation(state.lat, state.lon)
            if agl is not None:
                state.height = m.relative_alt*scale_relative_alt + state.home_height - agl
        elif mtype == 'ATTITUDE':
            state.roll, state.pitch, state.yaw = math.degrees(m.roll), math.degrees(m.pitch), math.degrees(m.yaw)
        elif mtype in ['GPS_RAW', 'GPS_RAW_INT']:
            if self.module('wp').wploader.count() > 0:
                home = self.module('wp').wploader.wp(0).x, self.module('wp').wploader.wp(0).y
            else:
                home = [self.master.field('HOME', c)*scale_latlon for c in ['lat', 'lon']]
            old = state.home_height # TODO TMP
            agl = state.elevation_model.GetElevation(*home)
            if agl is None:
                return
            state.home_height = agl

            # TODO TMP
            if state.home_height != old:
                # tridge said to get home pos from wploader,
                # but this is not the same as from master() below...!!
                # using master() gives the right coordinates
                # (i.e. matches GLOBAL_POSITION_INT coords, and $IMHOME in sim_arduplane.sh)
                # and wploader is a bit off
                print('home height changed from',old,'to',state.home_height)
        elif mtype == 'SERVO_OUTPUT_RAW':
            for (axis, attr) in [('ROLL', 'mount_roll'), ('TILT', 'mount_pitch'), ('PAN', 'mount_yaw')]:
                channel = int(self.get_mav_param('MNT_RC_IN_{0}'.format(axis), 0))
                if self.get_mav_param('MNT_STAB_{0}'.format(axis), 0) and channel:
                    # enabled stabilisation on this axis
                    # TODO just guessing that RC_IN_ROLL gives the servo number, but no idea if this is really the case
                    servo = 'servo{0}_raw'.format(channel)
                    centidegrees = self.scale_rc(getattr(m, servo),
                                            self.get_mav_param('MNT_ANGMIN_{0}'.format(axis[:3])),
                                            self.get_mav_param('MNT_ANGMAX_{0}'.format(axis[:3])),
                                            param='RC{0}'.format(channel))
                    setattr(state, attr, centidegrees*0.01)
            #state.mount_roll = min(max(-state.roll,-45),45)#TODO TMP
            #state.mount_yaw = min(max(-state.yaw,-45),45)#TODO TMP
            #state.mount_pitch = min(max(-state.pitch,-45),45)#TODO TMP
        else:
            return
        if self.mpstate.map: # if the map module is loaded, redraw polygon
            # get rid of the old polygon
            self.mpstate.map.add_object(mp_slipmap.SlipClearLayer('CameraView'))

            # camera view polygon determined by projecting corner pixels of the image onto the ground
            pixel_positions = [cuav_util.pixel_position(px[0],px[1], state.height, state.pitch+state.mount_pitch, state.roll+state.mount_roll, state.yaw+state.mount_yaw, state.camera_params) for px in [(0,0), (state.camera_params.xresolution,0), (state.camera_params.xresolution,state.camera_params.yresolution), (0,state.camera_params.yresolution)]]
            if any(pixel_position is None for pixel_position in pixel_positions):
                # at least one of the pixels is not on the ground
                # so it doesn't make sense to try to draw the polygon
                return
            gps_positions = [mp_util.gps_newpos(state.lat, state.lon, math.degrees(math.atan2(*pixel_position)), math.hypot(*pixel_position)) for pixel_position in pixel_positions]

            # draw new polygon
            self.mpstate.map.add_object(mp_slipmap.SlipPolygon('cameraview', gps_positions+[gps_positions[0]], # append first element to close polygon
                                                          layer='CameraView', linewidth=2, colour=state.col))

def init(mpstate):
    '''initialise module'''
    return CameraViewModule(mpstate)



'''
Cesium map module
Samuel Dudley
Jan 2016
'''
import os, json, time, sys, uuid

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

from pymavlink import mavutil
import threading
try:
    import Queue as queue
except ImportError:
    import queue

from MAVProxy.modules.mavproxy_cesium.app import cesium_web_server # the tornado web server

import webbrowser # open url's in browser window

from MAVProxy.modules.mavproxy_cesium.app.config import Configuration

        
class CesiumModule(mp_module.MPModule):
    def __init__(self, mpstate, **kwargs):
        super(CesiumModule, self).__init__(mpstate, "cesium", "Cesium map module", public = True)
        self.add_command('cesium', self.cmd_cesium, [""])
        self.data_stream = ['NAV_CONTROLLER_OUTPUT', 'VFR_HUD',
                            'ATTITUDE', 'GLOBAL_POSITION_INT',
                            'SYS_STATUS', 'MISSION_CURRENT',
                            'STATUSTEXT', 'FENCE_STATUS', 'WIND']
        
        # if a configuration path was passed when the module was loaded attempt to use it
        configuration_path = kwargs.get('configuration', None)
            
        self.config = Configuration(configuration_path)
        self.main_counter = 0
        
        self.message_queue = queue.Queue()
        
        self.wp_change_time = 0
        self.fence_change_time = 0
        self.rally_change_time = 0
        self.flightmode = None
        
        self.cesium_settings = mp_settings.MPSettings(
            [ ('openbrowser', bool, False),
              ('debug', bool, self.config.MODULE_DEBUG)])
        
        self.aircraft = {'lat':None, 'lon':None, 'alt_wgs84':None,
                         'roll':None, 'pitch':None, 'yaw':None}
        self.pos_target = {'lat':None, 'lon':None, 'alt_wgs84':None}
        self.fence = {}
        self.mission = {}
        
        self.server_thread = None
        self.start_server()
        
        if self.cesium_settings.openbrowser:
            self.open_display_in_browser()
    
    def start_server(self):
        if self.main_counter == 0:
            self.main_counter += 1
            self.server_thread = threading.Thread(target=cesium_web_server.main, args = (self.config, self.callback,))
            self.server_thread.daemon = True
            self.server_thread.start()
#             log.startLogging(sys.stdout)
            self.mpstate.console.writeln('MAVCesium display loaded at http://'+self.config.SERVER_INTERFACE+":"+self.config.SERVER_PORT+'/'+self.config.APP_PREFIX, fg='white', bg='blue')
        else:
            time.sleep(0.1)
        
    def stop_server(self):
        cesium_web_server.stop_tornado(self.config)
    
    def open_display_in_browser(self):
        if self.web_server_thread.isAlive():
            url = 'http://'+self.config.SERVER_INTERFACE+":"+self.config.SERVER_PORT+'/'+self.config.APP_PREFIX
            try:
                browser_controller = webbrowser.get('google-chrome')
                browser_controller.open_new_tab(url)
            except:
                webbrowser.open_new_tab(url)
                
    def callback(self, data):
        '''callback for data coming in from a websocket'''
        self.message_queue.put_nowait(data)
            
    def send_data(self, data, target = None):
        '''push json data to the browser via a websocket'''
        payload = json.dumps(data).encode('utf8')
        cesium_web_server.websocket_send_message(payload)
        # TODO: direct messages to individual websockets, e.g. new connections

    def cmd_cesium(self, args):
        '''cesium command parser'''
        usage = "usage: cesium <restart> <count> <set> (CESIUMSETTING)"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.cesium_settings.command(args[1:])
        elif args[0] == "count":
            print('%u connected' % int(len(cesium_web_server.live_web_sockets)))
        elif args[0] == "restart":
            self.restart()
        else:
            print(usage)
            
    def send_defines(self, target = None):
        '''get the current mav defines and send them'''
        miss_cmds = {}
        frame_enum = {0: "Abs", 3: "Rel", 10: "AGL"}
        
        # auto-generate the list of mission commands
        for cmd in mavutil.mavlink.enums['MAV_CMD']:
            enum = mavutil.mavlink.enums['MAV_CMD'][cmd]
            name = enum.name
            name = name.replace('MAV_CMD_','')
            if name == 'ENUM_END':
                continue
            miss_cmds[cmd] = name
        
        self.defines = {}
        self.defines['frame_enum'] = frame_enum
        self.defines['mission_commands'] = miss_cmds
        self.send_data({"defines":self.defines}, target = target)


    def send_fence(self):
        '''load and draw the fence in cesium'''
        self.fence = {}
        self.fence_points_to_send = self.mpstate.public_modules['fence'].fenceloader.points
        for point in self.fence_points_to_send:
            point_dict = point.to_dict()
            idx = point_dict['idx']
            del point_dict['idx']
            if idx != 0: # dont include the return location
                self.fence[idx] = point_dict
        self.send_data({"fence_data":self.fence})
            
    def send_mission(self):
        '''load and draw the mission in cesium'''
        self.mission = {}
        self.mission_points_to_send = self.mpstate.public_modules['wp'].wploader.wpoints
        for point in self.mission_points_to_send:
            point_dict = point.to_dict()
            seq = point_dict['seq']
            del point_dict['seq']
            self.mission[seq] = point_dict
        self.send_data({"mission_data":self.mission})
        
    def send_flightmode(self):
        self.send_data({"flightmode":self.master.flightmode})
        self.flightmode = self.master.flightmode
        
    def restart(self):
        '''restart the web server'''
        self.stop_server() 
        self.start_server()

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if self.master.flightmode != self.flightmode:
            self.send_flightmode()
        
        if m.get_type() == 'POSITION_TARGET_GLOBAL_INT':
            msg_dict = m.to_dict()
            self.pos_target['lat']= msg_dict['lat_int']
            self.pos_target['lon'] = msg_dict['lon_int']
            self.pos_target['alt_wgs84'] = msg_dict['alt']
             
            if None not in self.pos_target.values():
                self.send_data({"pos_target_data":self.pos_target})
        
        if m.get_type() in self.data_stream:
            msg_dict = m.to_dict()
            msg_dict['timestamp'] = m._timestamp
            self.send_data({'mav_data':msg_dict})
                   
                
        # if the waypoints have changed, redisplay
        last_wp_change = self.module('wp').wploader.last_change
        if self.wp_change_time != last_wp_change and abs(time.time() - last_wp_change) > 1:
            self.wp_change_time = last_wp_change
            self.send_mission()

            # this may have affected the landing lines from the rally points:
            self.rally_change_time = time.time()
    
        # if the fence has changed, redisplay
        if self.fence_change_time != self.module('fence').fenceloader.last_change:
            self.fence_change_time = self.module('fence').fenceloader.last_change
            self.send_fence()
                
    def idle_task(self):
        '''called on idle'''
        while not self.message_queue.empty():
            payload = self.message_queue.get_nowait()
            if self.cesium_settings.debug:
                print(payload)
                 
            if 'new_connection' in payload.keys():
                self.send_defines(target=payload['new_connection'])
                self.send_fence()
                self.send_mission()
                self.send_flightmode()
             
            elif 'mode_set' in payload.keys():
                self.mpstate.functions.process_stdin('%s' % (payload['mode_set']))
             
            elif 'wp_set' in payload.keys():
                self.mpstate.functions.process_stdin('wp set %u' % int(payload['wp_set']))
             
            elif 'wp_move' in payload.keys():
                self.mpstate.functions.process_stdin('wp move %u %f %f' % (
                                                                           int(payload['wp_move']['idx']),
                                                                           float(payload['wp_move']['lat']),
                                                                           float(payload['wp_move']['lon'])
                                                                           )
                                                     )
 
            elif 'wp_remove' in payload.keys():
                self.mpstate.functions.process_stdin('wp remove %u' % int(payload['wp_remove']))
             
            elif 'wp_list' in payload.keys():
                self.mpstate.functions.process_stdin('wp list')
                 
            elif 'fence_list' in payload.keys():
                self.mpstate.functions.process_stdin('fence list')
             
            else:
                pass
   
    def unload(self):
        '''unload module'''
        # override the unload method
        self.stop_server() 
        
def init(mpstate, **kwargs):
    '''initialise module'''
    return CesiumModule(mpstate, **kwargs)






#!/usr/bin/env python
'''mavlink file transfer support'''

import io
import time, os, sys
import struct
import random
from pymavlink import mavutil

try:
    # py2
    from StringIO import StringIO as SIO
except ImportError:
    # py3
    from io import BytesIO as SIO

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

# opcodes
OP_None = 0
OP_TerminateSession = 1
OP_ResetSessions = 2
OP_ListDirectory = 3
OP_OpenFileRO = 4
OP_ReadFile = 5
OP_CreateFile = 6
OP_WriteFile = 7
OP_RemoveFile = 8
OP_CreateDirectory = 9
OP_RemoveDirectory = 10
OP_OpenFileWO = 11
OP_TruncateFile = 12
OP_Rename = 13
OP_CalcFileCRC32 = 14
OP_BurstReadFile = 15
OP_Ack = 128
OP_Nack = 129

# error codes
ERR_None = 0
ERR_Fail = 1
ERR_FailErrno = 2
ERR_InvalidDataSize = 3
ERR_InvalidSession = 4
ERR_NoSessionsAvailable = 5
ERR_EndOfFile = 6
ERR_UnknownCommand = 7
ERR_FileExists = 8
ERR_FileProtected = 9
ERR_FileNotFound = 10

HDR_Len = 12
MAX_Payload = 239

class FTP_OP:
    def __init__(self, seq, session, opcode, size, req_opcode, burst_complete, offset, payload):
        self.seq = seq
        self.session = session
        self.opcode = opcode
        self.size = size
        self.req_opcode = req_opcode
        self.burst_complete = burst_complete
        self.offset = offset
        self.payload = payload

    def pack(self):
        '''pack message'''
        ret = struct.pack("<HBBBBBBI", self.seq, self.session, self.opcode, self.size, self.req_opcode, self.burst_complete, 0, self.offset)
        if self.payload is not None:
            ret += self.payload
        ret = bytearray(ret)
        return ret

    def __str__(self):
        plen = 0
        if self.payload is not None:
            plen = len(self.payload)
        ret = "OP seq:%u sess:%u opcode:%d req_opcode:%u size:%u bc:%u ofs:%u plen=%u" % (self.seq,
                                                                                          self.session,
                                                                                          self.opcode,
                                                                                          self.req_opcode,
                                                                                          self.size,
                                                                                          self.burst_complete,
                                                                                          self.offset,
                                                                                          plen)
        if plen > 0:
            ret += " [%u]" % self.payload[0]
        return ret

class WriteQueue:
    def __init__(self, ofs, size):
        self.ofs = ofs
        self.size = size
        self.last_send = 0

class FTPModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(FTPModule, self).__init__(mpstate, "ftp", public=True)
        self.add_command('ftp', self.cmd_ftp, "file transfer",
                         ["<list|get|rm|rmdir|rename|mkdir|crc|cancel|status>",
                          "set (FTPSETTING)",
                          "put (FILENAME) (FILENAME)"])
        self.ftp_settings = mp_settings.MPSettings(
            [('debug', int, 0),
             ('pkt_loss_tx', int, 0),
             ('pkt_loss_rx', int, 0),
             ('max_backlog', int, 5),
             ('burst_read_size', int, 110),
             ('write_size', int, 110),
             ('write_qsize', int, 5),
             ('retry_time', float, 0.5)])
        self.add_completion_function('(FTPSETTING)',
                                     self.ftp_settings.completion)
        self.seq = 0
        self.session = 0
        self.network = 0
        self.last_op = None
        self.fh = None
        self.filename = None
        self.callback = None
        self.callback_progress = None
        self.put_callback = None
        self.put_callback_progress = None
        self.total_size = 0
        self.read_gaps = []
        self.read_gap_times = {}
        self.last_gap_send = 0
        self.read_retries = 0
        self.read_total = 0
        self.duplicates = 0
        self.last_read = None
        self.last_burst_read = None
        self.op_start = None
        self.dir_offset = 0
        self.last_op_time = time.time()
        self.rtt = 0.5
        self.reached_eof = False
        self.backlog = 0
        self.burst_size = self.ftp_settings.burst_read_size
        self.write_list = None
        self.write_block_size = 0
        self.write_acks = 0
        self.write_total = 0
        self.write_file_size = 0
        self.write_idx = 0
        self.write_recv_idx = -1
        self.write_pending = 0
        self.write_last_send = None

    def cmd_ftp(self, args):
        '''FTP operations'''
        usage = "Usage: ftp <list|get|put|rm|rmdir|rename|mkdir|crc>"
        if len(args) < 1:
            print(usage)
            return
        if args[0] == 'list':
            self.cmd_list(args[1:])
        elif args[0] == "set":
            self.ftp_settings.command(args[1:])
        elif args[0] == 'get':
            self.cmd_get(args[1:])
        elif args[0] == 'put':
            self.cmd_put(args[1:])
        elif args[0] == 'rm':
            self.cmd_rm(args[1:])
        elif args[0] == 'rmdir':
            self.cmd_rmdir(args[1:])
        elif args[0] == 'rename':
            self.cmd_rename(args[1:])
        elif args[0] == 'mkdir':
            self.cmd_mkdir(args[1:])
        elif args[0] == 'crc':
            self.cmd_crc(args[1:])
        elif args[0] == 'status':
            self.cmd_status()
        elif args[0] == 'cancel':
            self.cmd_cancel()
        else:
            print(usage)

    def send(self, op):
        '''send a request'''
        op.seq = self.seq
        payload = op.pack()
        plen = len(payload)
        if plen < MAX_Payload + HDR_Len:
            payload.extend(bytearray([0]*((HDR_Len+MAX_Payload)-plen)))
        self.master.mav.file_transfer_protocol_send(self.network, self.target_system, self.target_component, payload)
        self.seq = (self.seq + 1) % 256
        self.last_op = op
        now = time.time()
        if self.ftp_settings.debug > 1:
            print("> %s dt=%.2f" % (op, now - self.last_op_time))
        self.last_op_time = time.time()

    def terminate_session(self):
        '''terminate current session'''
        self.send(FTP_OP(self.seq, self.session, OP_TerminateSession, 0, 0, 0, 0, None))
        self.fh = None
        self.filename = None
        self.write_list = None
        if self.callback is not None:
            # tell caller that the transfer failed
            self.callback(None)
            self.callback = None
        if self.put_callback is not None:
            # tell caller that the transfer failed
            self.put_callback(None)
            self.put_callback = None
        if self.put_callback_progress is not None:
            self.put_callback_progress(None)
            self.put_callback_progress = None
        self.read_gaps = []
        self.read_total = 0
        self.read_gap_times = {}
        self.last_read = None
        self.last_burst_read = None
        self.session = (self.session + 1) % 256
        self.reached_eof = False
        self.backlog = 0
        self.duplicates = 0
        if self.ftp_settings.debug > 0:
            print("Terminated session")

    def cmd_list(self, args):
        '''list files'''
        if len(args) > 0:
            dname = args[0]
        else:
            dname = '/'
        print("Listing %s" % dname)
        enc_dname = bytearray(dname, 'ascii')
        self.total_size = 0
        self.dir_offset = 0
        op = FTP_OP(self.seq, self.session, OP_ListDirectory, len(enc_dname), 0, 0, self.dir_offset, enc_dname)
        self.send(op)

    def handle_list_reply(self, op, m):
        '''handle OP_ListDirectory reply'''
        if op.opcode == OP_Ack:
            dentries = sorted(op.payload.split(b'\x00'))
            #print(dentries)
            for d in dentries:
                if len(d) == 0:
                    continue
                self.dir_offset += 1
                try:
                    if sys.version_info.major >= 3:
                        d = str(d, 'ascii')
                    else:
                        d = str(d)
                except Exception:
                    continue
                if d[0] == 'D':
                    print(" D %s" % d[1:])
                elif d[0] == 'F':
                    (name, size) = d[1:].split('\t')
                    size = int(size)
                    self.total_size += size
                    print("   %s\t%u" % (name, size))
                else:
                    print(d)
            # ask for more
            more = self.last_op
            more.offset = self.dir_offset
            self.send(more)
        elif op.opcode == OP_Nack and len(op.payload) == 1 and op.payload[0] == ERR_EndOfFile:
            print("Total size %.2f kByte" % (self.total_size / 1024.0))
            self.total_size = 0
        else:
            print('LIST: %s' % op)

    def cmd_get(self, args, callback=None, callback_progress=None):
        '''get file'''
        if len(args) == 0:
            print("Usage: get FILENAME <LOCALNAME>")
            return
        self.terminate_session()
        fname = args[0]
        if len(args) > 1:
            self.filename = args[1]
        else:
            self.filename = os.path.basename(fname)
        if callback is None or self.ftp_settings.debug > 1:
            print("Getting %s as %s" % (fname, self.filename))
        self.op_start = time.time()
        self.callback = callback
        self.callback_progress = callback_progress
        self.read_retries = 0
        self.duplicates = 0
        self.reached_eof = False
        self.burst_size = self.ftp_settings.burst_read_size
        if self.burst_size < 1:
            self.burst_size = 239
        elif self.burst_size > 239:
            self.burst_size = 239
        enc_fname = bytearray(fname, 'ascii')
        self.open_retries = 0
        op = FTP_OP(self.seq, self.session, OP_OpenFileRO, len(enc_fname), 0, 0, 0, enc_fname)
        self.send(op)

    def handle_open_RO_reply(self, op, m):
        '''handle OP_OpenFileRO reply'''
        if op.opcode == OP_Ack:
            if self.filename is None:
                return
            try:
                if self.callback is not None or self.filename == '-':
                    self.fh = SIO()
                else:
                    self.fh = open(self.filename, 'wb')
            except Exception as ex:
                print("Failed to open %s: %s" % (self.filename, ex))
                self.terminate_session()
                return
            read = FTP_OP(self.seq, self.session, OP_BurstReadFile, self.burst_size, 0, 0, 0, None)
            self.last_burst_read = time.time()
            self.send(read)
        else:
            if self.callback is None or self.ftp_settings.debug > 0:
                print("ftp open failed")
            self.terminate_session()

    def check_read_finished(self):
        '''check if download has completed'''
        if self.reached_eof and len(self.read_gaps) == 0:
            ofs = self.fh.tell()
            dt = time.time() - self.op_start
            rate = (ofs / dt) / 1024.0
            if self.callback is not None:
                self.fh.seek(0)
                self.callback(self.fh)
                self.callback = None
            elif self.filename == "-":
                self.fh.seek(0)
                if sys.version_info.major < 3:
                    print(self.fh.read())
                else:
                    print(self.fh.read().decode('utf-8'))
            else:
                print("Wrote %u bytes to %s in %.2fs %.1fkByte/s" % (ofs, self.filename, dt, rate))
            self.terminate_session()
            return True
        return False

    def write_payload(self, op):
        '''write payload from a read op'''
        self.fh.seek(op.offset)
        self.fh.write(op.payload)
        self.read_total += len(op.payload)
        if self.callback_progress is not None:
            self.callback_progress(self.fh, self.read_total)
    
    def handle_burst_read(self, op, m):
        '''handle OP_BurstReadFile reply'''
        if self.ftp_settings.pkt_loss_tx > 0:
            if random.uniform(0,100) < self.ftp_settings.pkt_loss_tx:
                if self.ftp_settings.debug > 0:
                    print("FTP: dropping TX")
                return
        if self.fh is None or self.filename is None:
            if op.session != self.session:
                # old session
                return
            print("FTP Unexpected burst read reply")
            print(op)
            return
        self.last_burst_read = time.time()
        size = len(op.payload)
        if size > self.burst_size:
            # this server doesn't handle the burst size argument
            self.burst_size = MAX_Payload
            if self.ftp_settings.debug > 0:
                print("Setting burst size to %u" % self.burst_size)
        if op.opcode == OP_Ack and self.fh is not None:
            ofs = self.fh.tell()
            if op.offset < ofs:
                # writing an earlier portion, possibly remove a gap
                gap = (op.offset, len(op.payload))
                if gap in self.read_gaps:
                    self.read_gaps.remove(gap)
                    self.read_gap_times.pop(gap)
                    if self.ftp_settings.debug > 0:
                        print("FTP: removed gap", gap, self.reached_eof, len(self.read_gaps))
                else:
                    if self.ftp_settings.debug > 0:
                        print("FTP: dup read reply at %u of len %u ofs=%u" % (op.offset, op.size, self.fh.tell()))
                    self.duplicates += 1
                    return
                self.write_payload(op)
                self.fh.seek(ofs)
                if self.check_read_finished():
                    return
            elif op.offset > ofs:
                # we have a gap
                gap = (ofs, op.offset-ofs)
                max_read = self.burst_size
                while True:
                    if gap[1] <= max_read:
                        self.read_gaps.append(gap)
                        self.read_gap_times[gap] = 0
                        break
                    g = (gap[0], max_read)
                    self.read_gaps.append(g)
                    self.read_gap_times[g] = 0
                    gap = (gap[0] + max_read, gap[1] - max_read)
                self.write_payload(op)
            else:
                self.write_payload(op)
            if op.burst_complete:
                if op.size > 0 and op.size < self.burst_size:
                    # a burst complete with non-zero size and less than burst packet size
                    # means EOF
                    if not self.reached_eof and self.ftp_settings.debug > 0:
                        print("EOF at %u with %u gaps t=%.2f" % (self.fh.tell(), len(self.read_gaps), time.time() - self.op_start))
                    self.reached_eof = True
                    if self.check_read_finished():
                        return
                    self.check_read_send()
                    return
                more = self.last_op
                more.offset = op.offset + op.size
                if self.ftp_settings.debug > 0:
                    print("FTP: burst continue at %u %u" % (more.offset, self.fh.tell()))
                self.send(more)
        elif op.opcode == OP_Nack:
            ecode = op.payload[0]
            if self.ftp_settings.debug > 0:
                print("FTP: burst nack: ", op)
            if ecode == ERR_EndOfFile or ecode == 0:
                if not self.reached_eof and op.offset > self.fh.tell():
                    # we lost the last part of the burst
                    if self.ftp_settings.debug > 0:
                        print("burst lost EOF %u %u" % (self.fh.tell(), op.offset))
                    return
                if not self.reached_eof and self.ftp_settings.debug > 0:
                    print("EOF at %u with %u gaps t=%.2f" % (self.fh.tell(), len(self.read_gaps), time.time() - self.op_start))
                self.reached_eof = True
                if self.check_read_finished():
                    return
                self.check_read_send()
            elif self.ftp_settings.debug > 0:
                print("FTP: burst Nack (ecode:%u): %s" % (ecode, op))
        else:
            print("FTP: burst error: %s" % op)

    def handle_reply_read(self, op, m):
        '''handle OP_ReadFile reply'''
        if self.fh is None or self.filename is None:
            if self.ftp_settings.debug > 0:
                print("FTP Unexpected read reply")
                print(op)
            return
        if self.backlog > 0:
            self.backlog -= 1
        if op.opcode == OP_Ack and self.fh is not None:
            gap = (op.offset, op.size)
            if gap in self.read_gaps:
                self.read_gaps.remove(gap)
                self.read_gap_times.pop(gap)
                ofs = self.fh.tell()
                self.write_payload(op)
                self.fh.seek(ofs)
                if self.ftp_settings.debug > 0:
                    print("FTP: removed gap", gap, self.reached_eof, len(self.read_gaps))
                if self.check_read_finished():
                    return
            elif op.size < self.burst_size:
                print("FTP: file size changed to %u" % op.offset+op.size)
                self.terminate_session()
            else:
                self.duplicates += 1
                if self.ftp_settings.debug > 0:
                    print("FTP: no gap read", gap, len(self.read_gaps))
        elif op.opcode == OP_Nack:
            print("Read failed with %u gaps" % len(self.read_gaps), str(op))
            self.terminate_session()
        self.check_read_send()
            
    def cmd_put(self, args, fh=None, callback=None, progress_callback=None):
        '''put file'''
        if len(args) == 0:
            print("Usage: put FILENAME <REMOTENAME>")
            return
        if self.write_list is not None:
            print("put already in progress")
            return
        fname = args[0]
        self.fh = fh
        if self.fh is None:
            try:
                self.fh = open(fname, 'rb')
            except Exception as ex:
                print("Failed to open %s: %s" % (fname, ex))
                return
        if len(args) > 1:
            self.filename = args[1]
        else:
            self.filename = os.path.basename(fname)
        if callback is None:
            print("Putting %s as %s" % (fname, self.filename))
        self.fh.seek(0,2)
        file_size = self.fh.tell()
        self.fh.seek(0)

        # setup write list
        self.write_block_size = self.ftp_settings.write_size
        self.write_file_size = file_size

        write_blockcount = file_size // self.write_block_size
        if file_size % self.write_block_size != 0:
            write_blockcount += 1

        self.write_list = set(range(write_blockcount))
        self.write_acks = 0
        self.write_total = write_blockcount
        self.write_idx = 0
        self.write_recv_idx = -1
        self.write_pending = 0
        self.write_last_send = None

        self.put_callback = callback
        self.put_callback_progress = progress_callback
        self.read_retries = 0
        self.op_start = time.time()
        enc_fname = bytearray(self.filename, 'ascii')
        op = FTP_OP(self.seq, self.session, OP_CreateFile, len(enc_fname), 0, 0, 0, enc_fname)
        self.send(op)

    def put_finished(self, flen):
        '''finish a put'''
        if self.put_callback_progress:
            self.put_callback_progress(1.0)
            self.put_callback_progress = None
        if self.put_callback is not None:
            self.put_callback(flen)
            self.put_callback = None
        else:
            print("Sent file of length ", flen)
        
    def handle_create_file_reply(self, op, m):
        '''handle OP_CreateFile reply'''
        if self.fh is None:
            self.terminate_session()
            return
        if op.opcode == OP_Ack:
            self.send_more_writes()
        else:
            print("Create failed")
            self.terminate_session()

    def send_more_writes(self):
        '''send some more writes'''
        if len(self.write_list) == 0:
            # all done
            self.put_finished(self.write_file_size)
            self.terminate_session()
            return

        now = time.time()
        if self.write_last_send is not None:
            if now - self.write_last_send > max(min(10*self.rtt, 1),0.2):
                # we seem to have lost a block of replies
                self.write_pending = max(0, self.write_pending-1)

        n = min(self.ftp_settings.write_qsize-self.write_pending, len(self.write_list))
        for i in range(n):
            # send in round-robin, skipping any that have been acked
            idx = self.write_idx
            while idx not in self.write_list:
                idx = (idx + 1) % self.write_total
            ofs = idx * self.write_block_size
            self.fh.seek(ofs)
            data = self.fh.read(self.write_block_size)
            write = FTP_OP(self.seq, self.session, OP_WriteFile, len(data), 0, 0, ofs, bytearray(data))
            self.send(write)
            self.write_idx = (idx + 1) % self.write_total
            self.write_pending += 1
            self.write_last_send = now

    def handle_write_reply(self, op, m):
        '''handle OP_WriteFile reply'''
        if self.fh is None:
            self.terminate_session()
            return
        if op.opcode != OP_Ack:
            print("Write failed")
            self.terminate_session()
            return

        # assume the FTP server processes the blocks sequentially. This means
        # when we receive an ack that any blocks between the last ack and this
        # one have been lost
        idx = op.offset // self.write_block_size
        count = (idx - self.write_recv_idx) % self.write_total

        self.write_pending = max(0, self.write_pending - count)
        self.write_recv_idx = idx
        self.write_list.discard(idx)
        self.write_acks += 1
        if self.put_callback_progress:
            self.put_callback_progress(self.write_acks/float(self.write_total))
        self.send_more_writes()

    def cmd_rm(self, args):
        '''remove file'''
        if len(args) == 0:
            print("Usage: rm FILENAME")
            return
        fname = args[0]
        print("Removing %s" % fname)
        enc_fname = bytearray(fname, 'ascii')
        op = FTP_OP(self.seq, self.session, OP_RemoveFile, len(enc_fname), 0, 0, 0, enc_fname)
        self.send(op)

    def cmd_rmdir(self, args):
        '''remove directory'''
        if len(args) == 0:
            print("Usage: rmdir FILENAME")
            return
        dname = args[0]
        print("Removing %s" % dname)
        enc_dname = bytearray(dname, 'ascii')
        op = FTP_OP(self.seq, self.session, OP_RemoveDirectory, len(enc_dname), 0, 0, 0, enc_dname)
        self.send(op)

    def handle_remove_reply(self, op, m):
        '''handle remove reply'''
        if op.opcode != OP_Ack:
            print("Remove failed %s" % op)

    def cmd_rename(self, args):
        '''rename file'''
        if len(args) < 2:
            print("Usage: rename OLDNAME NEWNAME")
            return
        name1 = args[0]
        name2 = args[1]
        print("Renaming %s to %s" % (name1, name2))
        enc_name1 = bytearray(name1, 'ascii')
        enc_name2 = bytearray(name2, 'ascii')
        enc_both = enc_name1 + b'\x00' + enc_name2
        op = FTP_OP(self.seq, self.session, OP_Rename, len(enc_both), 0, 0, 0, enc_both)
        self.send(op)

    def handle_rename_reply(self, op, m):
        '''handle rename reply'''
        if op.opcode != OP_Ack:
            print("Rename failed %s" % op)

    def cmd_mkdir(self, args):
        '''make directory'''
        if len(args) < 1:
            print("Usage: mkdir NAME")
            return
        name = args[0]
        print("Creating directory %s" % name)
        enc_name = bytearray(name, 'ascii')
        op = FTP_OP(self.seq, self.session, OP_CreateDirectory, len(enc_name), 0, 0, 0, enc_name)
        self.send(op)

    def handle_mkdir_reply(self, op, m):
        '''handle mkdir reply'''
        if op.opcode != OP_Ack:
            print("Create directory failed %s" % op)

    def cmd_crc(self, args):
        '''get crc'''
        if len(args) < 1:
            print("Usage: crc NAME")
            return
        name = args[0]
        self.filename = name
        self.op_start = time.time()
        print("Getting CRC for %s" % name)
        enc_name = bytearray(name, 'ascii')
        op = FTP_OP(self.seq, self.session, OP_CalcFileCRC32, len(enc_name), 0, 0, 0, bytearray(enc_name))
        self.send(op)

    def handle_crc_reply(self, op, m):
        '''handle crc reply'''
        if op.opcode == OP_Ack and op.size == 4:
            crc, = struct.unpack("<I", op.payload)
            now = time.time()
            print("crc: %s 0x%08x in %.1fs" % (self.filename, crc, now - self.op_start))
        else:
            print("crc failed %s" % op)

    def cmd_cancel(self):
        '''cancel any pending op'''
        self.terminate_session()

    def cmd_status(self):
        '''show status'''
        if self.fh is None:
            print("No transfer in progress")
        else:
            ofs = self.fh.tell()
            dt = time.time() - self.op_start
            rate = (ofs / dt) / 1024.0
            print("Transfer at offset %u with %u gaps %u retries %.1f kByte/sec" % (ofs, len(self.read_gaps), self.read_retries, rate))

    def op_parse(self, m):
        '''parse a FILE_TRANSFER_PROTOCOL msg'''
        hdr = bytearray(m.payload[0:12])
        (seq, session, opcode, size, req_opcode, burst_complete, pad, offset) = struct.unpack("<HBBBBBBI", hdr)
        payload = bytearray(m.payload[12:])[:size]
        return FTP_OP(seq, session, opcode, size, req_opcode, burst_complete, offset, payload)

    def mavlink_packet(self, m):
        '''handle a mavlink packet'''
        mtype = m.get_type()
        if mtype == "FILE_TRANSFER_PROTOCOL":
            op = self.op_parse(m)
            now = time.time()
            dt = now - self.last_op_time
            if self.ftp_settings.debug > 1:
                print("< %s dt=%.2f" % (op, dt))
            self.last_op_time = now
            if self.ftp_settings.pkt_loss_rx > 0:
                if random.uniform(0,100) < self.ftp_settings.pkt_loss_rx:
                    if self.ftp_settings.debug > 1:
                        print("FTP: dropping packet RX")
                    return

            if op.req_opcode == self.last_op.opcode and op.seq == (self.last_op.seq + 1) % 256:
                self.rtt = max(min(self.rtt, dt), 0.01)
            if op.req_opcode == OP_ListDirectory:
                self.handle_list_reply(op, m)
            elif op.req_opcode == OP_OpenFileRO:
                self.handle_open_RO_reply(op, m)
            elif op.req_opcode == OP_BurstReadFile:
                self.handle_burst_read(op, m)
            elif op.req_opcode == OP_TerminateSession:
                pass
            elif op.req_opcode == OP_CreateFile:
                self.handle_create_file_reply(op, m)
            elif op.req_opcode == OP_WriteFile:
                self.handle_write_reply(op, m)
            elif op.req_opcode in [OP_RemoveFile, OP_RemoveDirectory]:
                self.handle_remove_reply(op, m)
            elif op.req_opcode == OP_Rename:
                self.handle_rename_reply(op, m)
            elif op.req_opcode == OP_CreateDirectory:
                self.handle_mkdir_reply(op, m)
            elif op.req_opcode == OP_ReadFile:
                self.handle_reply_read(op, m)
            elif op.req_opcode == OP_CalcFileCRC32:
                self.handle_crc_reply(op, m)
            else:
                print('FTP Unknown %s' % str(op))

    def send_gap_read(self, g):
        '''send a read for a gap'''
        (offset, length) = g
        if self.ftp_settings.debug > 0:
            print("Gap read of %u at %u rem=%u blog=%u" % (length, offset, len(self.read_gaps), self.backlog))
        read = FTP_OP(self.seq, self.session, OP_ReadFile, length, 0, 0, offset, None)
        self.send(read)
        self.read_gaps.remove(g)
        self.read_gaps.append(g)
        self.last_gap_send = time.time()
        self.read_gap_times[g] = self.last_gap_send
        self.backlog += 1

    def check_read_send(self):
        '''see if we should send another gap read'''
        if len(self.read_gaps) == 0:
            return
        g = self.read_gaps[0]
        now = time.time()
        dt = now - self.read_gap_times[g]
        if not self.reached_eof:
            # send gap reads once
            for g in self.read_gap_times.keys():
                if self.read_gap_times[g] == 0:
                    self.send_gap_read(g)
            return
        if self.read_gap_times[g] > 0 and dt > self.ftp_settings.retry_time:
            if self.backlog > 0:
                self.backlog -= 1
            self.read_gap_times[g] = 0

        if self.read_gap_times[g] != 0:
            # still pending
            return
        if not self.reached_eof and self.backlog >= self.ftp_settings.max_backlog:
            # don't fill queue too far until we have got past the burst
            return
        if now - self.last_gap_send < 0.05:
            # don't send too fast
            return
        self.send_gap_read(g)

    def idle_task(self):
        '''check for file gaps and lost requests'''
        now = time.time()

        # see if we lost an open reply
        if self.op_start is not None and now - self.op_start > 1.0 and self.last_op.opcode == OP_OpenFileRO:
            self.op_start = now
            self.open_retries += 1
            if self.open_retries > 2:
                # fail the get
                self.op_start = None
                self.terminate_session()
                return
            if self.ftp_settings.debug > 0:
                print("FTP: retry open")
            send_op = self.last_op
            self.send(FTP_OP(self.seq, self.session, OP_TerminateSession, 0, 0, 0, 0, None))
            self.session = (self.session + 1) % 256
            send_op.session = self.session
            self.send(send_op)

        if len(self.read_gaps) == 0 and self.last_burst_read is None and self.write_list is None:
            return

        if self.fh is None:
            return

        # see if burst read has stalled
        if not self.reached_eof and self.last_burst_read is not None and now - self.last_burst_read > self.ftp_settings.retry_time:
            dt = now - self.last_burst_read
            self.last_burst_read = now
            if self.ftp_settings.debug > 0:
                print("Retry read at %u rtt=%.2f dt=%.2f" % (self.fh.tell(), self.rtt, dt))
            self.send(FTP_OP(self.seq, self.session, OP_BurstReadFile, self.burst_size, 0, 0, self.fh.tell(), None))
            self.read_retries += 1

        # see if we can fill gaps
        self.check_read_send()

        if self.write_list is not None:
            self.send_more_writes()

def init(mpstate):
    '''initialise module'''
    return FTPModule(mpstate)



#!/usr/bin/env python

from __future__ import print_function

'''
DataFlash Logging Module
June 2015

ArduPilot supports transmission of DataFlash logs over MavLink.

This module pokes the UAV to start sending logs, and stores them
in a local directory.

The relevant code in the ArduPilot code base can be found in
libraries/DataFlash/DataFlash_MAVLink.*
'''

import os
import os.path
from pymavlink import mavutil
import errno
import sys

from MAVProxy.modules.lib import mp_module
import time
from MAVProxy.modules.lib import mp_settings


class dataflash_logger(mp_module.MPModule):
    def __init__(self, mpstate):
        """Initialise module.  We start poking the UAV for messages after this
        is called"""
        super(dataflash_logger, self).__init__(
            mpstate,
            "dataflash_logger",
            "logging of mavlink dataflash messages"
        )
        self.sender = None
        self.stopped = False
        self.time_last_start_packet_sent = 0
        self.time_last_stop_packet_sent = 0
        self.dataflash_dir = self._dataflash_dir(mpstate)
        self.download = 0
        self.prev_download = 0
        self.last_status_time = time.time()
        self.last_seqno = 0
        self.missing_blocks = {}
        self.acking_blocks = {}
        self.blocks_to_ack_and_nack = []
        self.missing_found = 0
        self.abandoned = 0
        self.dropped = 0
        self.armed = False

        self.log_settings = mp_settings.MPSettings(
            [('verbose', bool, False),
             ('rotate_on_disarm', bool, False),
             ('df_target_system', int, 0),
             ('df_target_component', int, mavutil.mavlink.MAV_COMP_ID_LOG)]
        )
        self.add_command('dataflash_logger',
                         self.cmd_dataflash_logger,
                         "dataflash logging control",
                         ['status', 'start', 'stop', 'rotate', 'set (LOGSETTING)'])
        self.add_completion_function('(LOGSETTING)',
                                     self.log_settings.completion)

    def usage(self):
        '''show help on a command line options'''
        return "Usage: dataflash_logger <status|start|stop|set>"

    def cmd_dataflash_logger(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print(self.usage())
        elif args[0] == "status":
            print(self.status())
        elif args[0] == "stop":
            self.sender = None
            self.stopped = True
        elif args[0] == "start":
            self.stopped = False
        elif args[0] == "set":
            self.log_settings.command(args[1:])
        elif args[0] == "rotate":
            self.rotate_log()
        else:
            print(self.usage())

    def _dataflash_dir(self, mpstate):
        '''returns directory path to store DF logs in.  May be relative'''
        return mpstate.status.logdir

    def new_log_filepath(self):
        '''returns a filepath to a log which does not currently exist and is
        suitable for DF logging'''
        ll_filepath = os.path.join(self.dataflash_dir, 'LASTLOG.TXT')
        if (os.path.exists(ll_filepath) and os.stat(ll_filepath).st_size != 0):
            fh = open(ll_filepath, 'rb')
            log_cnt = int(fh.read()) + 1
            fh.close()
        else:
            log_cnt = 1

        self.lastlog_file = open(ll_filepath, 'w+b')
        if sys.version_info[0] >= 3:
            self.lastlog_file.write(str.encode(log_cnt.__str__()))
        else:
            self.lastlog_file.write(log_cnt.__str__())
        self.lastlog_file.close()

        return os.path.join(self.dataflash_dir, '%u.BIN' % (log_cnt,))

    def start_new_log(self):
        '''open a new dataflash log, reset state'''
        filename = self.new_log_filepath()

        self.last_seqno = 0
        self.logfile = open(filename, 'w+b')
        print("DFLogger: logging started (%s)" % (filename))
        self.prev_cnt = 0
        self.download = 0
        self.prev_download = 0
        self.last_idle_status_printed_time = time.time()
        self.last_status_time = time.time()
        self.missing_blocks = {}
        self.acking_blocks = {}
        self.blocks_to_ack_and_nack = []
        self.missing_found = 0
        self.abandoned = 0
        self.dropped = 0

    def status(self):
        '''returns information about module'''
        if self.download is None:
            return "Not started"

        transferred = self.download - self.prev_download
        self.prev_download = self.download
        now = time.time()
        interval = now - self.last_status_time
        self.last_status_time = now
        return("DFLogger: %(state)s Rate(%(interval)ds):%(rate).3fkB/s "
               "Block:%(block_cnt)d Missing:%(missing)d Fixed:%(fixed)d "
               "Abandoned:%(abandoned)d" %
               {"interval": interval,
                "rate": transferred/(interval*1000),
                "block_cnt": self.last_seqno,
                "missing": len(self.missing_blocks),
                "fixed": self.missing_found,
                "abandoned": self.abandoned,
                "state": "Inactive" if self.stopped else "Active"})

    def idle_print_status(self):
        '''print out statistics every 10 seconds from idle loop'''
        now = time.time()
        if (now - self.last_idle_status_printed_time) >= 10:
            print(self.status())
            self.last_idle_status_printed_time = now

    def idle_send_acks_and_nacks(self):
        '''Send packets to UAV in idle loop'''
        max_blocks_to_send = 10
        blocks_sent = 0
        i = 0
        now = time.time()
        while (i < len(self.blocks_to_ack_and_nack) and
               blocks_sent < max_blocks_to_send):
            # print("ACKLIST: %s" %
            # ([x[1] for x in self.blocks_to_ack_and_nack],))
            stuff = self.blocks_to_ack_and_nack[i]

            [master, block, status, first_sent, last_sent] = stuff
            if status == 1:
                # print("DFLogger: ACKing block (%d)" % (block,))
                mavstatus = mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_ACK
                (target_sys, target_comp) = self.sender
                self.master.mav.remote_log_block_status_send(target_sys,
                                                             target_comp,
                                                             block,
                                                             mavstatus)
                blocks_sent += 1
                del self.acking_blocks[block]
                del self.blocks_to_ack_and_nack[i]
                continue

            if block not in self.missing_blocks:
                # we've received this block now
                del self.blocks_to_ack_and_nack[i]
                continue

            # give up on packet if we have seen one with a much higher
            # number (or after 60 seconds):
            if (self.last_seqno - block > 200) or (now - first_sent > 60):
                if self.log_settings.verbose:
                    print("DFLogger: Abandoning block (%d)" % (block,))
                del self.blocks_to_ack_and_nack[i]
                del self.missing_blocks[block]
                self.abandoned += 1
                continue

            i += 1
            # only send each nack every-so-often:
            if last_sent is not None:
                if now - last_sent < 0.1:
                    continue

            if self.log_settings.verbose:
                print("DFLogger: Asking for block (%d)" % (block,))
            mavstatus = mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_NACK
            (target_sys, target_comp) = self.sender
            self.master.mav.remote_log_block_status_send(target_sys,
                                                         target_comp,
                                                         block,
                                                         mavstatus)
            blocks_sent += 1
            stuff[4] = now

    def idle_task_started(self):
        '''called in idle task only when logging is started'''
        isarmed = self.master.motors_armed()
        if self.armed != isarmed:
            self.armed = isarmed
            if not self.armed and self.log_settings.rotate_on_disarm:
                self.rotate_log()

        if self.log_settings.verbose:
            self.idle_print_status()
        self.idle_send_acks_and_nacks()

    def idle_task_not_started(self):
        '''called in idle task only when logging is not running'''
        if not self.stopped:
            self.tell_sender_to_start()

    def idle_task(self):
        '''called rapidly by mavproxy'''
        if self.sender is not None:
            self.idle_task_started()
        else:
            self.idle_task_not_started()

    def tell_sender_to_stop(self, m):
        '''send a stop packet (if we haven't sent one in the last second)'''
        now = time.time()
        if now - self.time_last_stop_packet_sent < 1:
            return
        if self.log_settings.verbose:
            print("DFLogger: Sending stop packet")
        self.time_last_stop_packet_sent = now
        self.master.mav.remote_log_block_status_send(
            m.get_srcSystem(),
            m.get_srcComponent(),
            mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_STOP,
            1)

    def tell_sender_to_start(self):
        '''send a start packet (if we haven't sent one in the last second)'''
        now = time.time()
        if now - self.time_last_start_packet_sent < 1:
            return
        self.time_last_start_packet_sent = now

        if self.log_settings.verbose:
            print("DFLogger: Sending start packet")

        target_sys = self.log_settings.df_target_system
        target_comp = self.log_settings.df_target_component
        self.master.mav.remote_log_block_status_send(
            target_sys,
            target_comp,
            mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_START,
            1)

    def rotate_log(self):
        '''send a start packet and rotate log'''
        now = time.time()
        self.time_last_start_packet_sent = now

        if self.log_settings.verbose:
            print("DFLogger: rotating")

        target_sys = self.log_settings.df_target_system
        target_comp = self.log_settings.df_target_component

        for i in range(3):
            # send 3 stop packets
            self.master.mav.remote_log_block_status_send(
                target_sys,
                target_comp,
                mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_STOP,
                1)
        
        self.master.mav.remote_log_block_status_send(
            target_sys,
            target_comp,
            mavutil.mavlink.MAV_REMOTE_LOG_DATA_BLOCK_START,
            1)
        self.start_new_log()

    def packet_is_for_me(self, m):
        '''returns true if this packet is appropriately addressed'''
        if m.target_system != self.master.mav.srcSystem:
            return False
        if m.target_component != self.master.mav.srcComponent:
            return False
        # if have a sender we can also check the source address:
        if self.sender is not None:
            if (m.get_srcSystem(), m.get_srcComponent()) != self.sender:
                return False
        return True

    def do_ack_block(self, seqno):
        if seqno in self.acking_blocks:
            # already acking this one; we probably sent
            # multiple nacks and received this one
            # multiple times
            return

        now = time.time()

        # ACK the block we just got:
        self.blocks_to_ack_and_nack.append([self.master, seqno, 1, now, None])
        self.acking_blocks[seqno] = 1

        # NACK any blocks we haven't seen and should have:
        if(seqno - self.last_seqno > 1):
            for block in range(self.last_seqno+1, seqno):
                if block not in self.missing_blocks and \
                   block not in self.acking_blocks:
                    self.missing_blocks[block] = 1
                    if self.log_settings.verbose:
                        print("DFLogger: setting %d for nacking" % (block,))
                    self.blocks_to_ack_and_nack.append(
                        [self.master, block, 0, now, None]
                    )
        # print("\nmissed blocks: ",self.missing_blocks)

    def mavlink_packet(self, m):
        '''handle mavlink packets'''
        if m.get_type() == 'REMOTE_LOG_DATA_BLOCK':
            if not self.packet_is_for_me(m):
                self.dropped += 1
                return

            if self.sender is None and m.seqno == 0:
                if self.log_settings.verbose:
                    print("DFLogger: Received data packet - starting new log")
                self.start_new_log()
                self.sender = (m.get_srcSystem(), m.get_srcComponent())

            if self.sender is None:
                # No connection right now, and this packet did not start one
                return

            if self.stopped:
                # send a stop packet @1Hz until the other end gets the idea:
                self.tell_sender_to_stop(m)
                return

            if self.sender is not None:
                size = len(m.data)
                data = bytearray(m.data[:size])
                ofs = size*(m.seqno)
                self.logfile.seek(ofs)
                self.logfile.write(data)

                if m.seqno in self.missing_blocks:
                    if self.log_settings.verbose:
                        print("DFLogger: Got missing block: %d" % (m.seqno,))
                    del self.missing_blocks[m.seqno]
                    self.missing_found += 1
                    self.blocks_to_ack_and_nack.append(
                        [self.master, m.seqno, 1, time.time(), None]
                    )
                    self.acking_blocks[m.seqno] = 1
                    # print("DFLogger: missing: %s" %
                    # (str(self.missing_blocks),))
                else:
                    self.do_ack_block(m.seqno)
                    if self.last_seqno < m.seqno:
                        self.last_seqno = m.seqno
                self.download += size


def init(mpstate):
    '''initialise module'''
    return dataflash_logger(mpstate)
